'''
Created on 22 aug. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import rasterio
from rasterio.io import DatasetReader
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class RTToFile(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, raster, ofilename):
        '''
        Constructor
        '''
        if not isinstance(raster, DatasetReader):
            raise IllegalArgumentTypeException(raster, "DatasetReader")
        self.raster = raster

        if not isinstance(ofilename, str):
            raise IllegalArgumentTypeException(
                ofilename, "output filename (str)")
        self.ofilename = ofilename

    def run(self):
        with rasterio.open(self.ofilename, "w", **self.raster.meta) as dst:
            if (1 == self.raster.count):
                dst.write(self.raster.read(1), indexes=1)
            else:
                # write each band individually
                for band in range(self.raster.count):
                    # write data, band # (starting from 1)
                    dst.write(self.raster.read(band+1), band+1)

        return self.raster


"""
from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos
from t4gpd.morph.STBBox import STBBox
from t4gpd.raster.RTClip import RTClip
from t4gpd.raster.STRasterize import STRasterize

buildings = GeoDataFrameDemos.districtGraslinInNantesBuildings()
roi = STBBox(buildings, buffDist=0).run()
raster1 = rasterio.open(
    "/home/tleduc/data/mnt_asc/mnt_nantes_2005_1m/MNT_L93_0354_6690.asc")
raster2 = RTClip(raster1, roi).run()

# raster3 = STRasterize(buildings, raster2, attr="HAUTEUR").run()
raster3 = STRasterize(buildings, raster2, attr=None).run()

RTToFile(raster3, "/tmp/raster3.tif").run()
"""



'''
Created on 5 avr. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''

from numpy import arange, array, ndarray
from pyvista import MultiBlock, UnstructuredGrid
from t4gpd.commons.GeoDataFrameLib import GeoDataFrameLib
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from vtkmodules.util.vtkConstants import VTK_POLYGON, VTK_POLY_LINE, VTK_TRIANGLE, \
    VTK_VERTEX


class ToUnstructuredGrid(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, gdfs, fieldname=None):
        '''
        Constructor
        '''
        if not isinstance(gdfs, (list, tuple, ndarray)):
            raise IllegalArgumentTypeException(gdfs, 'list or tuple of GeoDataFrames')
        assert GeoDataFrameLib.shareTheSameCrs(*gdfs), 'gdfs are expected to share the same crs!'
        self.gdfs = gdfs
        self.fieldname = fieldname

    def __fromGeomToUnstructuredGrids(self, geom):
        # https://vtk.org/wp-content/uploads/2015/04/file-formats.pdf
        if ('Point' == geom.geom_type):
            point = [[xyz[0], xyz[1], 0 if len(xyz) == 2 else xyz[2]] for xyz in geom.coords]
            cells = [1, 0]
            cells, celltypes, points = array(cells), array([VTK_VERTEX]), array(point)
            return [ UnstructuredGrid(cells, celltypes, points) ]

        elif ('LineString' == geom.geom_type):
            points = [[xyz[0], xyz[1], 0 if len(xyz) == 2 else xyz[2]] for xyz in geom.coords]
            npoints = len(points)
            cells = [npoints] + arange(npoints).tolist()
            cells, celltypes, points = array(cells), array([VTK_POLY_LINE]), array(points)
            return [ UnstructuredGrid(cells, celltypes, points) ]

        elif ('Polygon' == geom.geom_type):
            if (GeomLib.isConvex(geom) or GeomLib.isHoled(geom)):
                points = [[xyz[0], xyz[1], 0 if len(xyz) == 2 else xyz[2]] for xyz in geom.exterior.coords]
                npoints = len(points)
                cells = [npoints] + arange(npoints).tolist()
                cells, celltypes, points = array(cells), array([VTK_POLYGON]), array(points)
                return [ UnstructuredGrid(cells, celltypes, points) ]
            else:
                result = []
                triangles = GeomLib.fromPolygonToListOfTriangles(geom)
                for triangle in triangles:
                    points = [[xyz[0], xyz[1], 0 if len(xyz) == 2 else xyz[2]] for xyz in triangle.exterior.coords]
                    npoints = len(points)
                    cells = [npoints] + arange(npoints).tolist()
                    cells, celltypes, points = array(cells), array([VTK_TRIANGLE]), array(points)
                    result.append(UnstructuredGrid(cells, celltypes, points))
                return result

        raise Exception('Unreachable instruction!')

    def __fromGeoDataFrameToDictOfUnstructuredGrids(self, gdf, rank):
        gdf = gdf.explode(ignore_index=True)
        # print(gdf.columns)
        result = {}
        for idx, row in gdf.iterrows():
            geom = row.geometry
            partialPolyUgrids = self.__fromGeomToUnstructuredGrids(geom)

            # we can add some attributes to the current UnstructuredGrid
            if self.fieldname is not None:
                attrValue = row[self.fieldname] if self.fieldname in gdf else float('nan')
                for partialPolyUgrid in partialPolyUgrids:
                    partialPolyUgrid.cell_data[self.fieldname] = attrValue

            for i, partialPolyUgrid in enumerate(partialPolyUgrids):
                result[f'{rank}_{idx}_{i}'] = partialPolyUgrid
        return result

    def run(self):
        result = {}
        rank = 0
        for rank, gdf in enumerate(self.gdfs):
            result.update(self.__fromGeoDataFrameToDictOfUnstructuredGrids(gdf, rank))
        polyBlocks = MultiBlock(result)
        polyGrid = polyBlocks.combine()
        return polyGrid



'''
Created on 23 aug. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from rasterio.features import shapes
from rasterio.io import DatasetReader
from shapely.geometry import shape
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class RTVectorize(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, raster):
        '''
        Constructor
        '''
        if not isinstance(raster, DatasetReader):
            raise IllegalArgumentTypeException(raster, "DatasetReader")
        self.raster = raster

    @staticmethod
    def __vectorize(raster):
        pixels = raster.read(1, masked=True)
        shape_gen = ((shape(s), v)
                     for s, v in shapes(pixels, transform=raster.transform))
        gdf = GeoDataFrame(
            dict(zip(["geometry", "pix_value"], zip(*shape_gen))), crs=raster.crs)
        return gdf

    def run(self):
        return RTVectorize.__vectorize(self.raster)


"""
import matplotlib.pyplot as plt
from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos
from t4gpd.morph.STBBox import STBBox
from t4gpd.raster.RTClip import RTClip
from t4gpd.raster.RTLoad import RTLoad
from t4gpd.raster.STRasterize import STRasterize

dtmdir = "/home/tleduc/data/mnt_asc/mnt_nantes_2005_1m"
raster = RTLoad(f"{dtmdir}/MNT_L93_0354_6690.asc").run()

buildings = GeoDataFrameDemos.districtRoyaleInNantesBuildings()
roi = STBBox(buildings, buffDist=0).run()
dtm = RTClip(raster, roi).run()
raster = STRasterize(buildings, dtm, attr="HAUTEUR").run()

# plt.imshow(raster.read(1), cmap="BrBG")
# plt.show()

gdf = RTVectorize(raster).run()
gdf.drop(gdf[gdf.pix_value == 0].index, inplace=True)

fig, ax = plt.subplots(figsize=(1 * 8.26, 1 * 8.26))
buildings.boundary.plot(ax=ax, color="red", linewidth=1)
gdf.plot(ax=ax, column="pix_value", alpha=0.4)
plt.axis("off")
plt.tight_layout()
plt.show()

buildings.to_file("/tmp/0.shp")
gdf.to_file("/tmp/1.shp")
"""



'''
Created on 9 avr. 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame, overlay, sjoin_nearest
from shapely import MultiLineString
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class AbstractGridLib(object):
    '''
    classdocs
    '''

    def __init__(self, gdf, dx, dy=None, encode=True):
        '''
        Constructor
        '''
        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, 'GeoDataFrame')
        self.gdf = gdf
        self.dx = dx
        self.dy = dx if (dy is None) else dy
        self.encode = encode

    def distanceToNearestContour(self, grid):
        gdf = self.gdf.copy(deep=True)
        gdf.geometry = gdf.geometry.apply(
            lambda geom: MultiLineString(GeomLib.toListOfLineStrings(geom)))
        grid2 = sjoin_nearest(
            grid, gdf[["geometry"]], how="inner", distance_col="dist_to_ctr", max_distance=None)
        if ("index_right" not in gdf) and ("index_right" in grid2):
            grid2.drop(columns="index_right", inplace=True)
        grid2.drop_duplicates(
            subset=["gid"], keep="first", ignore_index=True, inplace=True)
        return grid2

    def grid(self):
        raise NotImplementedError('grid() must be overridden!')

    def __overlayWithGrid(self, grid, how="difference"):
        _result = grid.loc[:, ["gid", "geometry"]]
        _result.geometry = _result.geometry.apply(lambda geom: geom.centroid)
        _result = overlay(_result, self.gdf, how)
        _result = set(_result.gid.to_list())
        return _result

    def indoorGrid(self):
        _grid = self.indoorOutdoorGrid()
        _grid = _grid.loc[_grid[_grid.indoor == 1].index]
        _grid.reset_index(drop=True, inplace=True)
        return _grid

    def indoorOutdoorGrid(self):
        _grid = self.grid()
        _gids = self.__overlayWithGrid(_grid, how="difference")
        _grid["indoor"] = _grid.gid.apply(
            lambda gid: 0 if (gid in _gids) else 1)
        return _grid

    def outdoorGrid(self):
        _grid = self.indoorOutdoorGrid()
        _grid = _grid.loc[_grid[_grid.indoor == 0].index]
        _grid.reset_index(drop=True, inplace=True)
        return _grid



'''
Created on 22 aug. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import rasterio
from numpy import nan, seterr
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.raster.AbstractRasterGeoProcess import AbstractRasterGeoProcess


class RTNdvi(AbstractRasterGeoProcess):
    '''
    classdocs
    '''

    def __init__(self, filename, debug=False):
        '''
        Constructor
        '''
        if not isinstance(filename, str):
            raise IllegalArgumentTypeException(
                filename, "input filename (str)")
        self.filename = filename
        self.debug = debug

    @staticmethod
    def __ndvi(filename):
        '''
        NDVI is a spectral approach used to assess vegetation.

        The density of vegetation (NDVI) at a certain point of the image is equal
        to the difference in the intensities of reflected light in the red and 
        infrared range divided by the sum of these intensities.

        raster (Landsat 8, Collection 2 Level-2):
            Band 1 - Blue
            Band 2 - Green
            Band 3 - Red
            Band 4 - Near Infrared
        '''
        with rasterio.open(filename, mode="r", nodata=0) as raster:
            red = raster.read(3).astype(float)
            nir = raster.read(4).astype(float)

            # Allow NumPy division by zero
            seterr(divide="ignore", invalid="ignore")
            ndvi = (nir - red) / (nir + red)

            # Set pixels whose values are outside the NDVI range (-1, 1) to NaN
            # Likely due to errors in the Landsat imagery
            ndvi[ndvi > 1] = nan
            ndvi[ndvi < -1] = nan

            out_meta = raster.meta.copy()
            out_meta.update({
                "driver": "GTiff",
                "height": raster.height,
                "width": raster.width,
                "transform": raster.transform,
                "crs": raster.crs,
                "count": 1,
                "dtype": ndvi.dtype,
            })

        return out_meta, ndvi

    def run(self):
        out_meta, ndvi = RTNdvi.__ndvi(self.filename)
        result = self._write_and_load(ndvi, out_meta, self.debug)
        return result


"""
src = "/home/tleduc/data/landsat/LC08_01_044_034_LC08_L1GT_044034_20130330_20170310_01_T2_LC08_L1GT_044034_20130330_20170310_01_T2_B1.TIF"
r = RTNdvi(src).run()
print(type(r))
"""



'''
Created on 20 sep. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import sqrt, asarray
from shapely.geometry import LineString, MultiLineString, MultiPolygon, Point, Polygon
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class PointsDensifierLib3D(object):
    '''
    classdocs
    '''

    @staticmethod
    def __densifyByCurvilinearAbscissa(line, curvAbsc, blockid, contourid):
        if not isinstance(line, LineString):
            raise IllegalArgumentTypeException(line, 'LineString')

        listOfCoords = line.coords
        result, nodeid = list(), 0
        for i in range(1, len(listOfCoords)):
            a, b = listOfCoords[i - 1], listOfCoords[i]
            speedv = asarray([ b[0] - a[0], b[1] - a[1], b[2] - a[2] ])
            norm = sqrt(speedv[0] ** 2 + speedv[1] ** 2 + speedv[2] ** 2)
            if (0.0 < norm):
                speedv = speedv / norm

                for _curvAbsc in curvAbsc:
                    samplePoint = Point(a + _curvAbsc * norm * speedv)

                    result.append({'node_id': ArrayCoding.encode([blockid, contourid, nodeid]),
                                   'motion_dir': ArrayCoding.encode(speedv),
                                   'geometry': samplePoint})
                    nodeid += 1

        return result

    @staticmethod
    def densifyByCurvilinearAbscissa(geom, curvAbsc, blockid, contourid=0):
        assert GeomLib.is3D(geom), 'geom is expected to be 3D!'

        if isinstance(geom, LineString):
            result = PointsDensifierLib3D.__densifyByCurvilinearAbscissa(
                geom, curvAbsc, blockid, contourid)

        elif isinstance(geom, Polygon):
            result = PointsDensifierLib3D.__densifyByCurvilinearAbscissa(
                geom.exterior, curvAbsc, blockid, contourid)
            for g in geom.interiors:
                contourid += 1
                result += PointsDensifierLib3D.__densifyByCurvilinearAbscissa(
                    g, curvAbsc, blockid, contourid)

        elif isinstance(geom, MultiLineString) or isinstance(geom, MultiPolygon):
            result = list()
            for g in geom.geoms:
                result += PointsDensifierLib3D.densifyByCurvilinearAbscissa(
                    g, curvAbsc, blockid, contourid)
                contourid += 1

        else:
            result = []

        return result



'''
Created on 27 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.avidon.commons.Wh import Wh
from t4gpd.avidon.scenarios.CredocBasedScenario import CredocBasedScenario


class FullLockdown(CredocBasedScenario):
    '''
    classdocs
    '''
    def energyDemandOfITEquipment(self, row, dt):
        hour, weekday = dt.hour, dt.weekday()

        wh = (
            self._n_gateways(row) * Wh.gateway()
            +self._n_smartobjects(row) * Wh.smartobject()
            +self._n_smartspeakers(row) * Wh.smartspeaker()
            )

        if (self.is_weekend(dt)):
            wh += (
                self._n_computers(row) * Wh.computer() * (14 <= hour <= 20)
                +self._n_smartphones(row) * Wh.smartphone() * (hour in [9, 20])
                +self._n_tablets(row) * Wh.tablet() * ((9 <= hour <= 12) or (14 <= hour <= 20))
                +self._n_tvs(row) * Wh.tv() * (9 <= hour <= 23)
                )
        else:
            wh += (
                self._n_computers(row) * Wh.computer() * ((9 <= hour <= 12) or (14 <= hour <= 20))
                +self._n_smartphones(row) * Wh.smartphone() * (21 == hour)
                +self._n_tablets(row) * Wh.tablet() * ((7 <= hour <= 9) or (19 <= hour <= 21))
                +self._n_tvs(row) * Wh.tv() * ((7 <= hour <= 9) or (13 == hour) or (19 <= hour <= 22))
                )

        return wh



'''
Created on 12 oct. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.grid.GridLib import GridLib


class STAdaptativeGrid(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, gdf, dx, thresholds=None, indoor=False, intoPoint=False,
                 encode=True):
        '''
        Constructor
        '''
        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, "GeoDataFrame")
        self.gdf = gdf

        self.dx = dx
        self.thresholds = self.dx if thresholds is None else thresholds

        if indoor in [None, True, False, "both"]:
            self.indoor = indoor
        else:
            raise Exception(
                "Illegal argument: indoor must be chosen in [None, True, False, 'both']!")
        self.intoPoint = intoPoint
        self.encode = encode

    def run(self):
        for i, (dx, threshold) in enumerate(zip(self.dx, self.thresholds)):
            if i == 0:
                grid0 = GridLib.getGrid1(self.gdf, self.dx[0])
            else:
                grid0 = GridLib.getSubgrid1(self.gdf, grid0, dx, threshold)

        if self.indoor is None:
            pass
        elif ("both" == self.indoor):
            grid0 = GridLib.getIndoorOutdoorGrid(self.gdf, grid0)
        elif self.indoor:
            grid0 = GridLib.getIndoorGrid(self.gdf, grid0)
        else:
            grid0 = GridLib.getOutdoorGrid(self.gdf, grid0)

        if self.intoPoint:
            grid0.geometry = grid0.centroid

        return grid0


"""
import matplotlib.pyplot as plt
from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos

buildings = GeoDataFrameDemos.ensaNantesBuildings()
grid = STAdaptativeGrid(buildings, dx=[32, 16, 8, 4], indoor=None).run()

buildings.to_file("/tmp/buildings.shp")
grid.to_file("/tmp/grid.shp")

fig, ax = plt.subplots(figsize=(1*8.26, 1*8.26))
ax.set_title(f"{len(grid)} cells", fontsize=24)
buildings.plot(ax=ax, color="grey")
grid.boundary.plot(ax=ax, color="red", linewidth=0.5)
ax.axis("off")
plt.show()
plt.close(fig)
"""



'''
Created on 6 sep. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame, overlay
from t4gpd.commons.GeoDataFrameLib import GeoDataFrameLib
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class STHeightOfRoughness(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, masks, grid, elevationFieldName="HAUTEUR", outputFieldName="hre"):
        '''
        Constructor
        '''
        if not isinstance(masks, GeoDataFrame):
            raise IllegalArgumentTypeException(
                masks, "GeoDataFrame")
        if not elevationFieldName in masks:
            raise Exception(
                f"Illegal argument: {elevationFieldName} is not a relevant field name!")
        if not isinstance(grid, GeoDataFrame):
            raise IllegalArgumentTypeException(
                grid, "GeoDataFrame")
        if outputFieldName in grid:
            raise Exception(
                f"Illegal argument: {outputFieldName} is an existing column of the grid!")

        if not GeoDataFrameLib.shareTheSameCrs(masks, grid):
            raise Exception(
                "Illegal argument: masks and grid must share shames CRS!")

        self.masks = masks
        # CLEAN GEOMETRIES
        self.masks.geometry = self.masks.geometry.apply(
            lambda g: g.buffer(0))
        self.elevationFieldName = elevationFieldName
        self.grid = grid
        self.outputFieldName = outputFieldName

    @staticmethod
    def __overlay(masks, grid, elevationFieldName, outputFieldName):
        grid["__CELL_AREA__"] = grid.geometry.apply(lambda g: g.area)
        grid["__CELL_ID__"] = range(len(grid))
        masks2 = overlay(masks, grid, how="intersection", keep_geom_type=True)
        masks2["__VOLUME__"] = masks2.apply(
            lambda row: row[elevationFieldName]*row.geometry.area, axis=1)
        masks2 = masks2.dissolve(by="__CELL_ID__",
                                 aggfunc={"__CELL_AREA__": "first",
                                          "__VOLUME__": "sum"},
                                 as_index=False)
        masks2[outputFieldName] = masks2.apply(
            lambda row: row.__VOLUME__/row.__CELL_AREA__, axis=1)
        grid2 = grid.merge(
            masks2[["__CELL_ID__", outputFieldName]], on="__CELL_ID__")
        grid.drop(columns=["__CELL_AREA__", "__CELL_ID__"], inplace=True)
        grid2.drop(columns=["__CELL_AREA__", "__CELL_ID__"], inplace=True)
        return grid2

    def run(self):
        return STHeightOfRoughness.__overlay(self.masks, self.grid, self.elevationFieldName, self.outputFieldName)



'''
Created on 16 dec. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from scipy.sparse.csgraph import minimum_spanning_tree
from shapely.geometry import LineString, MultiLineString
from shapely.geometry import Point
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.graph.AbstractUrbanGraphLib import AbstractUrbanGraphLib

import networkx as nx


class MCALib(object):
    '''
    classdocs
    '''

    def __init__(self, roads, roi=None):
        '''
        Constructor
        '''
        if not isinstance(roads, GeoDataFrame):
            raise IllegalArgumentTypeException(roads, 'GeoDataFrame')

        if roi is None:
            _roads = roads
            geometries = [geom for geom in roads.geometry]
        else:
            ids = roads.sindex.intersection(roi)
            geometries = [roads.loc[_id].geometry for _id in ids]
            _roads = GeoDataFrame([{'geometry': geom} for geom in geometries], crs=roads.crs)

        self.roads = _roads
        self.graph, self.icVertices = MCALib.__buildGraph(geometries)

    @staticmethod
    def __addCoupleOfVertices(graph, first, last, linestring):
        graph.add_edge(first, last, weight=linestring.length, object={
            'geometry': linestring, 'weight':linestring.length})

    @staticmethod
    def __addEdge(graph, ciVertices, icVertices, linestring):
        multigraph = 0
        coords = linestring.coords
        if (2 <= len(coords)):
            first = AbstractUrbanGraphLib.addVertex(ciVertices, icVertices, coords[0])
            last = AbstractUrbanGraphLib.addVertex(ciVertices, icVertices, coords[-1])

            if (first in graph) and (last in graph[first]):
                multigraph = 1
                # Handle multigraph (two vertices may be connected by more than one edge)
                newEdges = GeomLib.cutsLineStringByCurvilinearDistance(linestring, linestring.length / 2.0)
                for _edge in newEdges:
                    MCALib.__addEdge(graph, ciVertices, icVertices, _edge)
            else:
                MCALib.__addCoupleOfVertices(graph, first, last, linestring)
                MCALib.__addCoupleOfVertices(graph, last, first, LineString(reversed(coords)))
        return multigraph

    @staticmethod
    def __buildGraph(geometries):
        graph = nx.DiGraph()
        ciVertices = dict()
        icVertices = dict()

        multigraph = 0
        for geom in geometries:
            if isinstance(geom, MultiLineString):
                for _geom in geom.geoms:
                    multigraph += MCALib.__addEdge(graph, ciVertices, icVertices, _geom)
            elif isinstance(geom, LineString):
                multigraph += MCALib.__addEdge(graph, ciVertices, icVertices, geom)
            # else:
                # raise IllegalArgumentTypeException(geom, 'LineString or MultiLineString')

        if 0 < multigraph:
            print('Handling of %d multigraph cases!' % multigraph)

        return graph, icVertices

    def __fromGraphToGeoDataFrame(self, centralities, label):
        rows = []
        for nodeIndex, centrality in centralities.items():
            rows.append({
                'geometry': Point((self.icVertices[nodeIndex])),
                label: centrality
                })
        return GeoDataFrame(rows, crs=self.roads.crs)

    def betweenness_centrality(self):
        centralities = nx.betweenness_centrality(self.graph, weight='weight')
        return self.__fromGraphToGeoDataFrame(centralities, 'betweenness')

    def closeness_centrality(self):
        centralities = nx.closeness_centrality(self.graph, distance='weight')
        return self.__fromGraphToGeoDataFrame(centralities, 'closeness')

    def degree_centrality(self):
        centralities = nx.degree_centrality(self.graph)
        return self.__fromGraphToGeoDataFrame(centralities, 'degree_c')

    def minimum_spanning_tree(self):
        matrix = nx.to_scipy_sparse_array(self.graph, weight='weight')
        # print(nx.to_pandas_adjacency(self.graph, weight='weight'))
        mst = minimum_spanning_tree(matrix)
        mst = mst.todense()
        raise Exception('TODO: MUST BE IMPLEMENTED!')
        return mst



'''
Created on 24 avr. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from os.path import basename, splitext


class WarnUtils(object):
    '''
    classdocs
    '''

    @staticmethod
    def format_Warning(message, category, filename, lineno, line=""):
        '''
        import warnings
        warnings.formatwarning = WarnUtils.format_Warning
        warnings.warn(msg)
        '''
        # return f"{basename(filename)}:{lineno}: {category.__name__}: {message}\n"
        return f"{filename}:{lineno}: {category.__name__}: {message}\n"

    @staticmethod
    def format_Warning_alt(message, category, filename, lineno, line=""):
        '''
        import warnings
        warnings.formatwarning = WarnUtils.format_Warning_alt
        warnings.warn(msg)
        '''
        filename = basename(splitext(filename)[0])
        return f"{filename}: {message}\n"



'''
Created on 9 dec. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''

from geopandas import GeoDataFrame
from numpy import arange, isnan
from pandas import DataFrame
from shapely.geometry import CAP_STYLE, LineString, Polygon
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException

import matplotlib.pyplot as plt


class TimelineTool(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, df, label='label', start='start', stop='stop',
                 color=None, fontsize=16, xrotation=0, oFile=None):
        '''
        Constructor
        '''
        if not isinstance(df, DataFrame):
            raise IllegalArgumentTypeException(df, 'DataFrame')
        self.df = df

        for fieldname in [label, start, stop]:
            if not fieldname in df:
                raise Exception('%s is not a relevant field name!' % fieldname)
        self.label = label
        self.start = start
        self.stop = stop

        if not (color is None or (color in df)):
            raise Exception('%s is not a relevant field name!' % color)
        self.color = color
        self.fontsize = fontsize
        self.xrotation = xrotation
        self.oFile = oFile

    def __build(self):
        self.df[self.stop] = list(zip(self.df[self.start], self.df[self.stop]))
        self.df[self.stop] = self.df[self.stop].apply(
            lambda t: t[0] + 1 if (isnan(t[1])) else t[1])
        self.df['y'] = -arange(1, len(self.df) + 1)

        x0, x1 = self.df[self.start].min(), self.df[self.stop].max()
        x0, x1 = x0 - 1, x1 + 1

        # ARROW GEODATAFRAME
        d = 0.35
        arrow = GeoDataFrame([
            {'geometry': LineString([ (x0, 0), (x1, 0) ])},
            {'geometry': LineString([ (x1 - d, d), (x1, 0), (x1 - d, -d) ])},
            ])
        arrow.geometry = arrow.geometry.apply(
            lambda g: g.buffer(0.05, cap_style=CAP_STYLE.flat))

        # RECT GEODATAFRAME
        dy = 0.45
        rect = GeoDataFrame([
            {
                'geometry': Polygon([
                    (row[self.start], row.y - dy),
                    (row[self.stop], row.y - dy),
                    (row[self.stop], row.y + dy),
                    (row[self.start], row.y + dy) ]),
                'color': None if (self.color is None) else row[self.color]
            }
            for _, row in self.df.iterrows()
            ])
        return arrow, rect

    def run(self):
        arrow, rect = self.__build()

        fig, ax = plt.subplots(figsize=(1.5 * 8.26, 1.5 * 8.26))

        arrow.plot(ax=ax, edgecolor='dimgrey', color='grey')
        if (self.color is None):
            rect.plot(ax=ax, edgecolor='black', color='grey')
        else:
            for color in rect[self.color].unique():
                _rect = rect[ rect[self.color] == color ]
                _rect.plot(ax=ax, edgecolor=color, color=color)

        # ax.yaxis.set_visible(False)
        for axis in ['top', 'bottom', 'left', 'right']:
            ax.spines[axis].set_linewidth(0.0)
        ax.tick_params(axis='x', labelsize=self.fontsize, rotation=self.xrotation)
        ax.set_yticks(self.df.y, self.df[self.label], fontsize=self.fontsize)
        ax.grid(color='lightgrey', linestyle=':', linewidth=0.7)

        if self.oFile is None:
            plt.show()
        else:
            plt.savefig(self.oFile, bbox_inches='tight')            
        plt.close(fig)

'''
from io import StringIO
from pandas import read_csv

_sio = StringIO("""label;start;stop;color
Abc;2003;;red
Def;2007;2012;green
Ghi;2014;;blue
""")
df = read_csv(_sio, sep=';')

# TimelineTool(df).run()
TimelineTool(df, color='color', oFile='/tmp/timeline.pdf').run()
'''



'''
Created on 27 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from random import choice, randrange


class Wh(object):
    '''
    classdocs
    '''

    @staticmethod
    def computer():
        # 50 Wh < laptop < 100 Wh
        return 75.0

    @staticmethod
    def gateway():
        # 150 kWh/an < box < 300 kWh/an
        # return randrange(17, 35)
        return 26.0

    @staticmethod
    def tablet():
        # 2 Wh < tablet < 6 Wh
        return 4.0

    @staticmethod
    def smartobject():
        return 1.0

    @staticmethod
    def smartphone():
        # 10 Wh once a day!
        return 10.0

    @staticmethod
    def smartspeaker():
        # GHome: 5 Wh
        return 5.0

    @staticmethod
    def tv():
        # 40 Wh (Oled) < TV < 250 Wh (Plasma)
        # return choice([40.0] * 2 + [60.0] * 10 + [100.0] * 10 + [250.0] * 2)
        return 100.0



'''
Created on 22 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import roll
from t4gpd.commons.Distances import EuclideanDistance

from t4gpd.commons.crossroads_identification.AbstractMethod import AbstractMethod


class MeanAngularityMethod(AbstractMethod):
    '''
    classdocs
    '''
    def __init__(self, maxRayLength, maxThresholdRatio=0.9):
        self.maxRayLength = maxRayLength
        self.maxThresholdRatio = maxThresholdRatio

    def attrName(self):
        return 'recId_avgA'

    def distance(self, signal, signalRef):
        dists = [EuclideanDistance.compute(roll(signal, offset), signalRef) for offset in range(len(signal))]
        return min(dists)

    def signature(self, raylens, isAPatternLayer):
        azims = self.__azimOffsets(raylens)
        return azims

    def __azimOffsets(self, raylens):
        aw = self.__maximaSequence(raylens)
        if (1 >= len(aw)):
            return [360]
        azimuths = sorted(aw.keys())
        _azimOffsets = []
        prev = -1
        for azim in azimuths:
            if (-1 < prev):
                _azimOffsets.append(azim - prev)
            prev = azim
        if (1 < len(azimuths)):
            _azimOffsets.append(360.0 + azimuths[0] - prev)
        return list(map(int, _azimOffsets))

    def __maximaSequence(self, raylens):
        nRays = len(raylens)

        aw = dict()
        azim, width = -1, -1
        threshold = self.maxThresholdRatio * self.maxRayLength
        for i in range(nRays):
            if (threshold < raylens[i]):
                if 0 == i:
                    j = -1
                    while (threshold < raylens[j]) and (-nRays < j):
                        j -= 1
                    if (-nRays == j):
                        return aw
                    azim = ((nRays + j) * 360.0) / nRays
                    width = 0
                    for _ in range(j, 1):
                        azim += 0.5
                        width += 1
                elif (-1 != width):
                    azim += 0.5
                    width += 1
                else:
                    azim = (i * 360.0) / nRays
                    width = 1
            else:
                if (-1 != width):
                    aw[azim % 360] = width
                    azim = width = -1
        return aw



'''
Created on 16 feb. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import datetime, timezone

from numpy import float64, ndarray
from t4gpd.avidon.commons.EnergyDemandCalculator import EnergyDemandCalculator
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class EnergyDemandOfITEquipment2(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, columns, dt, tz=timezone.utc, D=None, S=None, E=None, encode=False):
        '''
        Constructor
        '''
        self.calc = EnergyDemandCalculator(D, S, E)
        self.columns = columns

        if isinstance(dt, datetime):
            self.dt = [dt.replace(tzinfo=tz)] if dt.tzinfo is None else [dt]
        elif isinstance(dt, (list, ndarray, tuple)):
            self.dt = []
            for _dt in dt:
                if not isinstance(_dt, datetime):
                    raise IllegalArgumentTypeException(dt, 'datetime or list of datetimes')
                self.dt.append(_dt.replace(tzinfo=tz) if _dt.tzinfo is None else _dt)
        else:
            raise IllegalArgumentTypeException(dt, 'datetime or list of datetimes')

        self.encode = encode

    def runWithArgs(self, row):
        _C = row[self.columns].to_numpy(dtype=float64)

        whs, dts = [], []
        for _dt in self.dt:
            _wh = self.calc.energyDemandOfITEquipment(_C, _dt)
            dts.append(_dt)
            whs.append(_wh)

        if (1 == len(self.dt)):
            return { 'IT_in_Wh': whs[0], 'datetime': dts[0] }

        if self.encode:
            return { 'IT_in_Wh': ArrayCoding.encode(whs), 'datetime': ArrayCoding.encode(dts) }

        return { 'IT_in_Wh': whs, 'datetime': dts }



'''
Created on 27 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.avidon.scenarios.AbstractScenario import AbstractScenario


class CredocBasedScenario(AbstractScenario):
    '''
    classdocs
    '''

    def _n_computers(self, row):
        # ~ (CREDOC, 2019, p. 42) - how to integrate data from Table 7, p. 43
        return (0.91 * row['Ind_11_17']
            +0.82 * row['Ind_18_24']
            +0.73 * row['Ind_25_39']
            +0.83 * row['Ind_40_54']
            +0.76 * row['Ind_55_64']
            +0.58 * row['Ind_65_79'])

    def _n_dsl(self, row):
        # ~ (CREDOC, 2019, p. 80)
        return (0.66 * self._n_gateways(row))

    def _n_gateways(self, row):
        # ~ (CREDOC, 2019, p. 76)
        return (0.86 * row['Men'])

    def _n_smartobjects(self, row):
        # ~ (CREDOC, 2019, p. 54)
        return (0.31 * row['Ind_11_17']
            +0.25 * row['Ind_18_24']
            +0.21 * row['Ind_25_39']
            +0.17 * row['Ind_40_54']
            +0.09 * row['Ind_55_64']
            +0.05 * row['Ind_65_79'])

    def _n_smartphones(self, row):
        # ~ (CREDOC, 2019, p. 35)
        return (0.86 * row['Ind_11_17']
            +0.98 * row['Ind_18_24']
            +0.95 * row['Ind_25_39']
            +0.80 * row['Ind_40_54']
            +0.62 * row['Ind_55_64']
            +0.44 * row['Ind_65_79'] 
            +0.0 * row['Ind_80p'])

    def _n_smartspeakers(self, row):
        # ~ (CREDOC, 2019, p. 49)
        return (0.15 * row['Ind_11_17']
            +0.14 * row['Ind_18_24']
            +0.10 * row['Ind_25_39']
            +0.09 * row['Ind_40_54']
            +0.07 * row['Ind_55_64']
            +0.05 * row['Ind_65_79'])

    def _n_tablets(self, row):
        # ~ (CREDOC, 2019, p. 46)
        return (0.37 * row['Ind_11_17']
            +0.36 * row['Ind_18_24']
            +0.49 * row['Ind_25_39']
            +0.45 * row['Ind_40_54']
            +0.47 * row['Ind_55_64']
            +0.29 * row['Ind_65_79'])

    def _n_tth(self, row):
        # ~ (CREDOC, 2019, p. 80)
        return (0.29 * self._n_gateways(row))

    def _n_tvs(self, row):
        # ~ Equipement des menages - 2018
        # ~ https://www.insee.fr/fr/statistiques/4277714
        return (0.792 * row['Ind_18_24']
            +0.931 * row['Ind_25_39']
            +0.953 * row['Ind_40_54']
            +0.979 * row['Ind_55_64']
            +0.979 * row['Ind_65_79'])



'''
Created on 8 dec. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import date, datetime, timedelta, timezone

from t4gpd.commons.LatLonLib import LatLonLib
from t4gpd.commons.sun.SunLib import SunLib


class DaylightLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def generate(dt, dtDelta=timedelta(minutes=15), gdf=LatLonLib.NANTES,
                 model='pysolar'):
        sunlib = SunLib(gdf, model)
        sunrise, sunset = sunlib.getSunrise(dt), sunlib.getSunset(dt)
        dts = []
        dt = sunrise + dtDelta
        while dt < sunset:
            dts.append(dt)
            dt += dtDelta
        return dts

"""
dt = date(2020, 12, 21)
dt = datetime(2020, 12, 21)
dts = DaylightLib.generate(dt, dtDelta=timedelta(hours=1))
for dt in dts:
    print(dt)
"""



'''
Created on 15 feb. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''

from numpy import dot, matmul, multiply, vectorize
from t4gpd.avidon.commons.AbstractEnergyDemand import AbstractEnergyDemand


class EnergyDemandCalculator(AbstractEnergyDemand):
    '''
    classdocs
    '''

    def __init__(self, D=None, S=None, E=None):
        '''
        Constructor
        '''
        super().__init__(D, S, E)

    def __eval(self, ch):
        h, w = self.dt.hour, self.dt.weekday()
        delta = lambda t: self._unitImpulse(t)
        rect = lambda t0, t1, t: self._indicatorFunction(t0, t1, t)
        # snd = lambda x: self._randomVariates(x)
        snd = lambda x: x
        return eval(ch) if isinstance(ch, str) else ch 

    def energyDemandOfITEquipment(self, C, dt):
        self.dt = dt
        _eval_vector = vectorize(self.__eval)
        D = _eval_vector(self.D) / 100.0
        S = _eval_vector(self.S) / 100.0
        E = _eval_vector(self.E)
        res = dot(matmul(D, multiply(S, C)), E)
        return res



'''
Created on 15 sept. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import warnings
from zoneinfo import ZoneInfo

from pandas import read_csv
from t4gpd.commons.WarnUtils import WarnUtils
from t4gpd.io.AbstractReader import AbstractReader


class KestrelReader(AbstractReader):
    '''
    classdocs

    https://kestrelinstruments.com/kestrel-5400-heat-stress-tracker
    '''

    def __init__(self, inputFile, tzinfo="Europe/Paris"):
        '''
        Constructor
        '''
        warnings.formatwarning = WarnUtils.format_Warning
        self.inputFile = inputFile
        self.tzinfo = ZoneInfo(tzinfo)

    def __getVersion(self):
        V1 = "FORMATTED DATE-TIME,Direction – True,Wind Speed,Crosswind Speed,Headwind Speed,Temperature,Globe Temperature,Wind Chill,Relative Humidity,Heat Stress Index,Dew Point,Psychro Wet Bulb Temperature,Station Pressure,Barometric Pressure,Altitude,Density Altitude,NA Wet Bulb Temperature,WBGT"
        V2 = "FORMATTED DATE_TIME,Temperature,Wet Bulb Temp,Globe Temperature,Relative Humidity,Barometric Pressure,Altitude,Station Pressure,Wind Speed,Heat Index,Dew Point,Density Altitude,Crosswind,Headwind,Compass Magnetic Direction,NWB Temp,Compass True Direction,Thermal Work Limit,Wet Bulb Globe Temperature,Wind Chill"
        with KestrelReader.opener(self.inputFile) as f:
            for nline, line in enumerate(f.readlines(), start=1):
                if (4 == nline):
                    if line.startswith(V1):
                        return 1
                    elif line.startswith(V2):
                        return 2
                    raise NotImplementedError(f'\n\n\t*** KestrelReader:: unknown file format version! ***')

    def __readV1(self):
        FIELD_NAMES = [
            "timestamp", "WindDirection", "WindSpeed", "CrosswindSpeed", "HeadwindSpeed",
            "Tair", "GlobeTemperature", "WindChill", "RH", "HeatStressIdx", "DewPoint",
            "PsychroWetBulbTemp", "StationPressure", "BarometricPressure", "Altitude",
            "DensityAltitude", "NAWetBulbTemperature", "WBGT", "TWL", "DirectionMag"] 

        df = read_csv(KestrelReader.opener(self.inputFile), decimal=",",
                      header=None, names=FIELD_NAMES, na_values="***", sep=",",
                      skiprows=5, parse_dates=[0], usecols=range(20))

        dfUnits = read_csv(KestrelReader.opener(self.inputFile), header=None,
                           names=FIELD_NAMES, nrows=1, sep=",", skiprows=4,
                           usecols=range(20))

        if (("km/h" == dfUnits.WindSpeed.squeeze().lower()) or 
            ("km/h" == dfUnits.CrosswindSpeed.squeeze().lower()) or 
            ("km/h" == dfUnits.HeadwindSpeed.squeeze().lower())):
            warnings.warn("Speed in km/h")
            # CONVERT km/h INTO m/s
            for fieldname in ["WindSpeed", "CrosswindSpeed", "HeadwindSpeed"]:
                if ("km/h" == dfUnits[fieldname].squeeze().lower()):
                    df[fieldname] /= 3.6

        df["sensorFamily"] = "Kestrel-5400"

        return df

    def __readV2(self):
        FIELD_NAMES = [
            "timestamp", "Tair", "WetBulbTemperature", "GlobeTemperature", "RH",
            "BarometricPressure", "Altitude", "StationPressure", "WindSpeed", "HeatStressIdx",
            "DewPoint", "DensityAltitude", "CrosswindSpeed", "HeadwindSpeed", "CompassMD",
            "NAWetBulbTemperature", "WindDirection", "TWL", "WBGT", "WindChill", "DataType"] 

        df = read_csv(KestrelReader.opener(self.inputFile), decimal=".",
                      header=None, names=FIELD_NAMES, na_values="--", sep=",",
                      skiprows=5, parse_dates=[0], usecols=range(21))

        dfUnits = read_csv(KestrelReader.opener(self.inputFile), header=None,
                           names=FIELD_NAMES, nrows=1, sep=",", skiprows=4,
                           usecols=range(21))

        if (("km/h" == dfUnits.WindSpeed.squeeze().lower()) or 
            ("km/h" == dfUnits.CrosswindSpeed.squeeze().lower()) or 
            ("km/h" == dfUnits.HeadwindSpeed.squeeze().lower())):
            warnings.warn("Speed in km/h")
            # CONVERT km/h INTO m/s
            for fieldname in ["WindSpeed", "CrosswindSpeed", "HeadwindSpeed"]:
                if ("km/h" == dfUnits[fieldname].squeeze().lower()):
                    df[fieldname] /= 3.6

        df["sensorFamily"] = "Kestrel-5400"

        return df

    def run(self):
        version = self.__getVersion()
        if (1 == version):
            warnings.warn("KestrelReader (v1)")
            df = self.__readV1()
        elif (2 == version):
            df = self.__readV2()
        else:
            raise Exception("Unreachable instruction!")

        df.timestamp = df.timestamp.apply(lambda dt: dt.replace(tzinfo=self.tzinfo))
        if not df.timestamp.is_monotonic_increasing:
            msg = f"Timestamps in {self.inputFile} are not monotonically increasing"
            warnings.warn(msg)

        return df

"""
inputFile = "/home/tleduc/prj/coolscapes-tl/dev/data/05-Quai-de-plantes-Nantes/20200518/Kestrel/export_HEAT2414248_2020_5_18_16_44_43.csv"
inputFile = "/home/tleduc/prj/coolscapes-tl/dev/data/06-Carriere-misery-Nantes/20200720/Kestrel/export_HEAT2414248_2020_7_20_16_32_54.csv"
df = KestrelReader(inputFile).run()
print(df.head(5))

inputFile = "/home/tleduc/prj/coolscapes-tl/dev/data/08-La-defense-Paris/20210904/Kestrel/HEAT_-_2414248_9_4_21_18_31_40.csv"
inputFile = "/home/tleduc/prj/nm-ilots-frais/terrain/220713/HEAT_-_2414248_7_13_22_18_36_20.csv"
inputFile = "/home/tleduc/prj/nm-ilots-frais/terrain/220711/HEAT_-_2414248_7_11_22_18_55_00.csv"
df = KestrelReader(inputFile).run()
print(df.head(5))
"""



'''
Created on 24 jul. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from numpy import arctan2, asarray, cos, linspace, pi, sin
from shapely import Polygon
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.commons.GeoDataFrameLib import GeoDataFrameLib
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.RayCasting4Lib import RayCasting4Lib


class STSkyMap25D(GeoProcess):
    '''
    classdocs
    '''
    PIDIV2 = 0.5 * pi

    def __init__(self, buildings, viewpoints, nRays=64, rayLength=100.0,
                 elevationFieldname="HAUTEUR", h0=0.0, size=4.0, epsilon=1e-2,
                 projectionName="Stereographic", withIndices=False, withAngles=False,
                 encode=False):
        '''
        Constructor
        '''
        if not isinstance(buildings, GeoDataFrame):
            raise IllegalArgumentTypeException(
                buildings, "buildings GeoDataFrame")
        if not elevationFieldname in buildings:
            raise Exception(
                f"{elevationFieldname} is not a relevant field name!")
        if not isinstance(viewpoints, GeoDataFrame):
            raise IllegalArgumentTypeException(
                viewpoints, "viewpoints GeoDataFrame")

        if not GeoDataFrameLib.shareTheSameCrs(buildings, viewpoints):
            raise Exception(
                "Illegal argument: buildings and viewpoints must share shames CRS!")

        self.buildings = buildings
        self.elevationFieldname = elevationFieldname
        # CLEAN GEOMETRIES
        self.buildings.geometry = self.buildings.geometry.apply(
            lambda g: g.buffer(0))
        self.buildings.geometry = self.buildings.apply(
            lambda row: row.geometry if row.geometry.has_z else GeomLib.forceZCoordinateToZ0(
                row.geometry, row[self.elevationFieldname]),
            axis=1
        )

        self.nRays = nRays
        self.rayLength = rayLength
        self.rays = RayCasting4Lib.get25DPanopticRaysGeoDataFrame(
            viewpoints, rayLength, nRays, h0)

        self.size = size
        self.epsilon = epsilon
        if not projectionName in ["Stereographic"]:
            raise IllegalArgumentTypeException(
                projectionName, "spherical projection as 'Stereographic'")
        self.proj = self.__stereographic

        self.withIndices = withIndices
        self.withAngles = withAngles
        self.encode = encode

    def __stereographic(self, lat, lon):
        radius = (self.size * cos(lat)) / (1.0 + sin(lat))
        return (radius * cos(lon), radius * sin(lon))

    def __angles(self, heights, widths, ray_ids):
        lats = [arctan2(h, w) for h, w in zip(heights, widths)]
        if (self.nRays != len(ray_ids)):
            ray_ids = {ray_id: lats[i] for i, ray_id in enumerate(ray_ids)}
            lats = [
                ray_ids[i] if i in ray_ids else self.PIDIV2 for i in range(self.nRays)]
        return lats

    def __buildSkyMap(self, viewpoint, lats, lons):
        viewpoint = viewpoint.centroid
        pnodes = [self.proj(lats[i], lons[i])
                  for i in range(len(lats))]
        pnodes = [(viewpoint.x + pnode[0], viewpoint.y + pnode[1])
                  for pnode in pnodes]
        return Polygon(viewpoint.buffer(self.size + self.epsilon).exterior.coords, [pnodes])

    def run(self):
        smapRaysField = RayCasting4Lib.multipleRayCast25D(
            self.buildings, self.rays, self.nRays, self.rayLength, self.elevationFieldname, self.withIndices, h0=0.0)

        if (0 < len(smapRaysField)):
            lons = linspace(0, 2*pi, self.nRays, endpoint=False)
            smapRaysField["angles"] = smapRaysField.apply(
                lambda row: self.__angles(
                    row.__RAY_DELTA_ALT__, row.__RAY_LEN__, row.__RAY_ID__),
                axis=1
            )
            smapRaysField.geometry = smapRaysField.apply(lambda row: self.__buildSkyMap(
                row.viewpoint, row.angles, lons), axis=1)
            smapRaysField.angles = smapRaysField.angles.apply(
                lambda v: (asarray(v) * 180)/pi)

        fields = ["__RAY_ID__"]
        for field in ["__RAY_LEN__", "__RAY_ALT__", "__RAY_DELTA_ALT__"]:
            if field in smapRaysField.columns:
                fields.append(field)

        if not self.withAngles:
            fields += ["angles"]

        smapRaysField.drop(columns=fields, inplace=True)

        if self.encode:
            smapRaysField.viewpoint = smapRaysField.viewpoint.apply(
                lambda vp: vp.wkt)
            if "angles" in smapRaysField:
                smapRaysField.angles = smapRaysField.angles.apply(
                    lambda a: ArrayCoding.encode(a))

        return smapRaysField


"""
import matplotlib.pyplot as plt
from shapely import Point
from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos

buildings = GeoDataFrameDemos.districtRoyaleInNantesBuildings()
pts = [ Point((355119.8, 6689339.2)), Point((355143.0, 6689359.4)),
	Point((355113.4, 6689397.9)) ]
sensors = GeoDataFrame([{"gid": i, "geometry": p} for i, p in enumerate(pts)],
	crs=buildings.crs)
# sensors.geometry = sensors.geometry.apply(lambda g: g.buffer(5))

skymaps = STSkyMap25D(buildings, sensors, withIndices=True).run()

print(skymaps)
fig, ax = plt.subplots(figsize=(8.26, 8.26))
buildings.plot(ax=ax, color="grey")
sensors.plot(ax=ax, color="red")
skymaps.plot(ax=ax, color="blue")
plt.show()
"""



'''
Created on 11 oct. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import re

from fiona import listlayers
from geopandas import read_file
from pandas import concat
from t4gpd.commons.Epsilon import Epsilon
from t4gpd.commons.GeoProcess import GeoProcess


class TracksWaypointsReader(GeoProcess):
    '''
    classdocs
    '''
    RE1 = re.compile(r'^track(\d*)$')
    RE2 = re.compile(r'^waypoints(\d*)$')

    def __init__(self, inputFile):
        '''
        Constructor
        '''
        self.inputFile = inputFile

    def run(self):
        static, tracks, waypoints = None, [], []
        layernames = sorted(listlayers(self.inputFile))
        for layer in layernames:
            if layer.startswith('static'):
                static = read_file(self.inputFile, layer=layer)

            elif layer.startswith('track'):
                num = int(self.RE1.search(layer).group(1))
                df = read_file(self.inputFile, layer=layer)
                df['id'] = num
                df = df[['id', 'geometry']].copy(deep=True)
                tracks.append(df)

            elif layer.startswith('waypoints'):
                num = int(self.RE2.search(layer).group(1))
                df = read_file(self.inputFile, layer=layer)
                df = df[['id', 'geometry']].copy(deep=True)
                idMin, idMax = df.id.min(), df.id.max()
                df['minTagName'], df['maxTagName'] = idMin, idMax
                assert (num == idMin // 100) and (num == idMax // 100), f'inconsistent ids in the waypoints{num} layer'
                df.sort_values(by='id', inplace=True)
                waypoints.append(df)

        if (len(tracks) != len(waypoints)):
            raise Exception("The number of tracks does not equal the number of waypoints layers!")

        for i in range(len(tracks)):
            _trackId, _trackGeom, _waypoints = tracks[i].id.squeeze(), tracks[i].geometry.squeeze(), waypoints[i]
            _waypoints['curv_absc'] = _waypoints.geometry.apply(
                lambda g: _trackGeom.project(g, normalized=True))

            if not _waypoints.curv_absc.is_monotonic_increasing:
                raise Exception(f"The abscissa of the \
waypoints should be increasing. Check the direction of track #{_trackId}, which should \
be consistent with the direction of the waypoints.")

            _waypoints['delta_curv_absc'] = _waypoints.curv_absc.diff(periods=1)
            _waypoints.delta_curv_absc = _waypoints.delta_curv_absc.shift(periods=-1)

        assert Epsilon.isZero(_waypoints.curv_absc.min()), f'Track {_trackId}: The min. curv. absc. must be equal to 0'
        assert Epsilon.equals(1.0, _waypoints.curv_absc.max()), f'Track {_trackId}: The max. curv. absc. must be equal to 1'

        tracks = concat(tracks, ignore_index=True)
        waypoints = concat(waypoints, ignore_index=True)
        return static, tracks, waypoints



'''
Created on 9 janv. 2023

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from numpy import arcsin, asarray, cos, pi, round, sin, sqrt, vstack, where
from numpy.random import default_rng, uniform
from shapely.geometry import Point, Polygon
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.GeomLib3D import GeomLib3D


class RandomPointPicking(object):
    '''
    classdocs
    '''

    @staticmethod
    def _simpleTriangulation(ngon):
        coords = ngon.exterior.coords[:-1]
        n = len(coords)
        resultat = []
        for i, ipp in zip(range(1, n - 1), range(2, n)):
            resultat.append(Polygon([coords[0], coords[i], coords[ipp]]))
        return resultat
    
    @staticmethod
    def convexPolygonPointPicking(ngon, npts):
        rng = default_rng()
        # INSPIRED BY:
        # https://blogs.sas.com/content/iml/2020/10/21/random-points-in-polygon.html
        assert isinstance(ngon, Polygon), 'ngon is expected to be a Shapely Polygon!'
        assert GeomLib.isConvex(ngon), 'ngon is expected to be convex!'
        tris = RandomPointPicking._simpleTriangulation(ngon)
        # areas = [tri.area for tri in tris]
        areas = asarray([GeomLib3D.getArea(tri) for tri in tris])
        ratios = areas / areas.sum() 
        listOfNpts = round(npts * ratios).astype(int)

        choices = [0] * len(tris) + [1] * npts
        listOfNpts[listOfNpts == 0] = rng.choice(choices, size=len(listOfNpts[listOfNpts == 0]))

        resultat = []
        for tri, _npts in zip(tris, listOfNpts):
            resultat.append(RandomPointPicking.trianglePointPicking(tri, _npts))
        return vstack(resultat)

    @staticmethod
    def trianglePointPicking(tri, npts):
        # INSPIRED BY:
        # https://blogs.sas.com/content/iml/2020/10/19/random-points-in-triangle.html
        assert isinstance(tri, Polygon), 'tri is expected to be a Shapely Polygon!'
        assert tri.is_valid, 'tri is expected to be a VALID Shapely Polygon!'
        assert (4 == len(tri.exterior.coords)), 'tri is expected to be a triangle!'
        a, b, c = list(tri.exterior.coords[:-1])
        a = [a[0], a[1], 0 if (2 == len(a)) else a[2]]
        b = [b[0], b[1], 0 if (2 == len(b)) else b[2]]
        c = [c[0], c[1], 0 if (2 == len(c)) else c[2]]
        ab = asarray([b[0] - a[0], b[1] - a[1], b[2] - a[2]])
        ac = asarray([c[0] - a[0], c[1] - a[1], c[2] - a[2]])

        # GENERATE RANDOM UNIFORM VARIABLES
        u1, u2 = uniform(low=0, high=1, size=[2, npts])
        idx = where(u1 + u2 > 1)
        u1[idx], u2[idx] = 1 - u1[idx], 1 - u2[idx]

        # GENERATE XYZ
        x = a[0] + u1 * ab[0] + u2 * ac[0]
        y = a[1] + u1 * ab[1] + u2 * ac[1]
        z = a[2] + u1 * ab[2] + u2 * ac[2]
        return vstack([x, y, z]).T 

    @staticmethod
    def unitDiskPointPicking(npts):
        # INSPIRED BY:
        # https://mathworld.wolfram.com/DiskPointPicking.html
        r, theta = uniform(low=0, high=1, size=[2, npts])
        r, theta = sqrt(r), 2 * pi * theta
        x, y = r * cos(theta), r * sin(theta)
        return vstack([x, y]).T 

    @staticmethod
    def unitSpherePointPicking(npts):
        # INSPIRED BY:
        # https://mathworld.wolfram.com/SpherePointPicking.html
        u, v = uniform(low=0, high=1, size=[2, npts])
        # LONGITUDE, LATITUDE
        lon, lat = 2 * pi * u, arcsin(2 * v - 1)
        # FROM LONGITUDE AND LATITUDE TO CARTESIAN COORDINATES
        cosLat = cos(lat)
        x, y, z = cos(lon) * cosLat, sin(lon) * cosLat, sin(lat)
        return vstack([x, y, z]).T 

    @staticmethod
    def toGeoDataFrame(XYZ):
        rows = []
        for gid, xyz in enumerate(XYZ, start=1):
            rows.append({ 'gid': gid, 'geometry': Point(xyz) })
        return GeoDataFrame(rows)



'''
Created on 11 oct. 2022

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from numpy import isnan, nan
from t4gpd.commons.GeoDataFrameLib import GeoDataFrameLib
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class SnapImuOnTrackUsingWaypoints(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, dfImu, tracks, waypoints):
        '''
        Constructor
        '''
        if not isinstance(dfImu, GeoDataFrame):
            raise IllegalArgumentTypeException(dfImu, "GeoDataFrame")
        for fieldname in ["timestamp", "TagName"]:
            if fieldname not in dfImu:
                raise Exception(f"{fieldname} is not a relevant field name!")
        assert dfImu.timestamp.is_monotonic_increasing, "dfImu.timestamp must be increasing!"
        assert dfImu.timestamp.is_unique, "dfImu.timestamp must be strictly increasing!"
        self.dfImu = dfImu

        if not isinstance(tracks, GeoDataFrame):
            raise IllegalArgumentTypeException(tracks, "GeoDataFrame")
        if "id" not in tracks:
            raise Exception(f"'id' is not a relevant field name!")
        self.tracks = tracks

        if not isinstance(waypoints, GeoDataFrame):
            raise IllegalArgumentTypeException(waypoints, "GeoDataFrame")
        if "id" not in waypoints:
            raise Exception(f"'id' is not a relevant field name!")
        self.waypoints = waypoints

        assert GeoDataFrameLib.shareTheSameCrs(dfImu, tracks), "dfImu and tracks must share the same crs!"
        assert GeoDataFrameLib.shareTheSameCrs(tracks, waypoints), "tracks and waypoints must share the same crs!"
        self.crs = tracks.crs

    def run(self):
        maxTagNames = self.waypoints.maxTagName.unique()
        imu = self.dfImu.copy(deep=True)
        imu.to_crs(self.crs, inplace=True)

        # ==============================================================
        # 1st STEP: IDENTIFY THE TRACK TO PROJECT ON
        imu["track"], imu["curv_absc"], imu["n_curv_absc"] = None, None, None

        prev, cnt1, rows, cnts = nan, nan, [], []
        for cnt2, (_, row) in enumerate(imu.iterrows()):
            curr = row.TagName
            if isnan(prev):
                prev, cnt1 = (nan, nan) if isnan(curr) else (curr, 0)
                row.track, row.curv_absc = prev, cnt1
                if not isnan(curr):
                    cnts.append(cnt2)

            else:
                if isnan(curr):
                    cnts.append(cnt2)
                    cnt1 += 1
                    row.track, row.curv_absc = prev, cnt1
                else:

                    row.track, row.curv_absc = curr, 0
                    row.n_curv_absc = cnt1 + 1
                    for _cnt in cnts:
                        rows[_cnt].n_curv_absc = cnt1 + 1
                    cnts = []

                    if curr in maxTagNames:
                        prev, cnt1 = nan, nan
                    else:
                        prev, cnt1 = curr, 0

            rows.append(row)
        imu = GeoDataFrame(rows, crs=self.crs)

        # ==============================================================
        # 2nd STEP: ASSESS THE CURVILINEAR ABSCISSA OF THE PROJECTED POINT
        wpDict = self.waypoints.set_index("id").to_dict(orient="index")

        imu.curv_absc = imu.apply(lambda row: None
                                  if isnan(row.track)
                                  else wpDict[row.track]["curv_absc"] + (row.curv_absc * wpDict[row.track]["delta_curv_absc"]) / row.n_curv_absc,
                                  axis=1)
        for tn in maxTagNames:
            imu.loc[ imu[imu.TagName == tn].index, "curv_absc" ] = 1.0

        # ==============================================================
        # 3rd STEP: ASSESS THE COORDINATES OF THE PROJECTED POINT
        trDict = {row.id: row.geometry for _, row in self.tracks.iterrows()}

        imu.track = imu.track.apply(lambda v: v if isnan(v) else int(v // 100))
        imu["snap_geometry"] = imu.apply(lambda row: None
                                         if isnan(row.track)
                                         else trDict[row.track].interpolate(row.curv_absc, normalized=True),
                                         axis=1)
 
        # ==============================================================
        # 4th STEP: ASSESS THE DISTANCE (DRIFT) BETWEEN THE GNSS POINT AND 
        # ITS CORRESPONDING PROJECTED POINT
        imu.rename(columns={"geometry": "gnss_geom", "snap_geometry": "geometry"}, inplace=True)
        imu.drop(columns=["n_curv_absc"], inplace=True)

        imu["drift"] = imu.apply(lambda row: nan 
                                 if ((row.gnss_geom is None) or (row.geometry is None))
                                 else row.gnss_geom.distance(row.geometry),
                                 axis=1)
        imu.gnss_geom = imu.gnss_geom.apply(lambda g: g.wkt)
        imu.set_crs(self.crs, inplace=True)

        imu.drop(index=imu[imu.geometry.isna()].index, inplace=True)
        imu.reset_index(drop=True, inplace=True)

        return imu

'''
import matplotlib.pyplot as plt
from t4gpd.picoclim.MetrologicalCampaignReader import MetrologicalCampaignReader

# dirName = "/home/tleduc/prj/nm-ilots-frais/terrain/220711"
dirName = "/home/tleduc/prj/nm-ilots-frais/terrain/220713"
static, tracks, waypoints, dfImu, dfMob, dfStat1, dfStat2, dfMeteoFr = MetrologicalCampaignReader(dirName).run()
dfImu.to_crs(crs=tracks.crs, inplace=True)

imu = SnapImuOnTrackUsingWaypoints(dfImu, tracks, waypoints).run()

# imu.drop(columns=["timestamp", "X", "Y", "Z", "Distance", "degree", "latitude",
#                   "longitude", "GpsAccuracy", "indoor_outdoor_flag"], inplace=True)
# imu.to_csv("/home/tleduc/prj/nm-ilots-frais/a.csv", index=False)
# imu.to_file("/tmp/imu.gpkg")
imu.plot()
plt.show()

'''



'''
Created on 24 jul. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import matplotlib.pyplot as plt
from geopandas import GeoDataFrame, overlay
from numpy import arange, asarray, cos, exp, pi, sin, stack, zeros
from scipy.stats import kurtosis, skew
from shapely import equals, LineString, MultiLineString, Point, Polygon
from shapely.affinity import translate
from shapely.ops import nearest_points
from t4gpd.commons.Entropy import Entropy
from t4gpd.commons.GeoDataFrameLib import GeoDataFrameLib
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.GeomLib3D import GeomLib3D
from t4gpd.commons.SVFLib import SVFLib


class RayCasting4Lib(object):
    '''
    classdocs
    '''
    @staticmethod
    def get2DPanopticRaysGeoDataFrame(sensors, rayLength=100.0, nRays=64):
        angles = ((2.0 * pi) / nRays) * arange(nRays)
        shootingDirs = rayLength * stack([cos(angles), sin(angles)], axis=1)
        shootingDirs = MultiLineString(
            [LineString([(0, 0), xy]) for xy in shootingDirs])

        rays2D = sensors.copy(deep=True)
        rays2D["viewpoint"] = rays2D.geometry
        rays2D.geometry = rays2D.geometry.apply(
            lambda geom: geom.centroid)
        rays2D.geometry = rays2D.geometry.apply(
            lambda geom: translate(shootingDirs, xoff=geom.x, yoff=geom.y))
        rays2D = rays2D.explode(index_parts=True)
        rays2D = rays2D.reset_index(names=["__VPT_ID__", "__RAY_ID__"])
        return rays2D

    @staticmethod
    def get25DPanopticRaysGeoDataFrame(sensors, rayLength=100.0, nRays=64, h0=0.0):
        rays25D = RayCasting4Lib.get2DPanopticRaysGeoDataFrame(
            sensors, rayLength, nRays)
        rays25D.geometry = rays25D.apply(
            lambda row: GeomLib.forceZCoordinateToZ0(
                row.geometry, z0=GeomLib3D.centroid(
                    row.viewpoint).z if row.viewpoint.has_z else h0), axis=1)
        rays25D.viewpoint = rays25D.viewpoint.apply(
            lambda vp: GeomLib.forceZCoordinateToZ0(
                vp, z0=GeomLib3D.centroid(vp).z if vp.has_z else h0))
        return rays25D

    @staticmethod
    def __keepTheAnchoredSegment(sensor, segments):
        epsilon = 1e-3
        if isinstance(segments, (Point, LineString)):
            if (epsilon > sensor.distance(segments)):
                return segments
        else:
            for segment in segments.geoms:
                if (epsilon > sensor.distance(segment)):
                    return segment
        return LineString([])

    @staticmethod
    def __buildIsovist(nRays, viewpoint, ray_ids, mls):
        if not GeomLib.isMultipart(mls):
            return Polygon([])

        _ctrPts1 = [ls.coords[-1] for ls in mls.geoms]
        if (nRays == len(_ctrPts1)):
            return Polygon(_ctrPts1)

        viewpoint = viewpoint.coords[0][0:2]
        assert len(ray_ids) == len(_ctrPts1)
        assert ray_ids == sorted(ray_ids)
        _ctrPts2 = []
        for i, ray_id in enumerate(ray_ids):
            if (0 == i):
                if (0 != ray_id):
                    _ctrPts2.append(viewpoint)
                _ctrPts2.append(_ctrPts1[i])
            else:
                if (ray_ids[i-1]+1 < ray_id):
                    _ctrPts2.append(viewpoint)
                _ctrPts2.append(_ctrPts1[i])
        return Polygon(_ctrPts2)

    @staticmethod
    def multipleRayCast2D(buildings, rays, withIndices):
        if not GeoDataFrameLib.shareTheSameCrs(buildings, rays):
            raise Exception(
                "Illegal argument: buildings and rays must share shames CRS!")

        nRays = len(rays)
        rays = overlay(rays, buildings, how="difference")
        rays.geometry = rays.apply(lambda row: RayCasting4Lib.__keepTheAnchoredSegment(
            row.viewpoint, GeomLib.removeZCoordinate(row.geometry)), axis=1)
        rays = rays.loc[rays[~rays.geometry.apply(
            lambda geom: geom.is_empty)].index, :]
        isovRaysField = rays.dissolve(
            by="__VPT_ID__", as_index=False, aggfunc=list)
        isovRaysField.drop(columns=["__VPT_ID__"], inplace=True)
        for fieldname in isovRaysField.columns:
            if fieldname not in ["geometry", "__RAY_ID__"]:
                isovRaysField[fieldname] = isovRaysField[fieldname].apply(
                    lambda t: t[0])

        if (0 < len(isovRaysField)):
            isovRaysField["__ISOV_GEOM__"] = isovRaysField.apply(lambda row: RayCasting4Lib.__buildIsovist(
                nRays, row.viewpoint, row.__RAY_ID__, row.geometry), axis=1)

        if withIndices:
            isovRaysField = RayCasting4Lib.addIndicesToIsovRaysField2D(
                nRays, isovRaysField)

        isovField = isovRaysField.copy(deep=True)
        if (0 < len(isovField)):
            isovField.geometry = isovField.__ISOV_GEOM__

        isovRaysField.drop(
            columns=["__RAY_ID__", "__ISOV_GEOM__"], inplace=True)
        isovField.drop(columns=["__RAY_ID__", "__ISOV_GEOM__"], inplace=True)

        return isovRaysField, isovField

    @staticmethod
    def __from2DRaysToRayLengths(nRays, ray_ids, mls):
        _raylens = asarray(
            [ray.length for ray in GeomLib.toListOfLineStrings(mls)])

        if (nRays == len(_raylens)):
            return _raylens

        raylens = zeros(nRays)
        for i, ray_id in enumerate(ray_ids):
            raylens[ray_id] = _raylens[i]
        return raylens

    @staticmethod
    def __add2DIndices(isovRaysField, precision=1.0, base=exp(1)):
        isovRaysField["w_mean"] = isovRaysField.__RAY_LEN__.apply(
            lambda raylens: float(raylens.mean()))
        isovRaysField["w_std"] = isovRaysField.__RAY_LEN__.apply(
            lambda raylens: float(raylens.std()))
        isovRaysField["w_kurtosis"] = isovRaysField.__RAY_LEN__.apply(
            lambda raylens: float(kurtosis(raylens, fisher=True, bias=True)))
        isovRaysField["w_skew"] = isovRaysField.__RAY_LEN__.apply(
            lambda raylens: float(skew(raylens)))
        isovRaysField["w_entropy"] = isovRaysField.__RAY_LEN__.apply(
            lambda raylens: Entropy.createFromDoubleValuesArray(
                raylens, precision).h(base)
        )
        return isovRaysField

    @staticmethod
    def addIndicesToIsovRaysField2D(nRays, isovRaysField, precision=1.0, base=exp(1)):
        isovRaysField["__RAY_LEN__"] = isovRaysField.apply(
            lambda row: RayCasting4Lib.__from2DRaysToRayLengths(nRays, row.__RAY_ID__, row.geometry), axis=1)

        isovRaysField = RayCasting4Lib.__add2DIndices(
            isovRaysField, precision, base)

        isovRaysField["__ISOV_CENTRE__"] = isovRaysField.__ISOV_GEOM__.apply(
            lambda geom: geom.centroid)

        isovRaysField["vect_drift"] = isovRaysField.apply(
            lambda row: None if row.__ISOV_CENTRE__.is_empty else LineString([
                row.viewpoint.coords[0][0:2], row.__ISOV_CENTRE__.coords[0]]), axis=1)
        isovRaysField["drift"] = isovRaysField.vect_drift.apply(
            lambda v: None if v is None else v.length)

        isovRaysField.drop(
            columns=["__RAY_LEN__", "__ISOV_CENTRE__"], inplace=True)
        return isovRaysField

    @staticmethod
    def __build_LineString(pA, pB):
        if (equals(pA, pB)) or (pA.is_empty) or (pB.is_empty):
            return LineString([])
        if (pA.has_z == pB.has_z):
            return LineString([pA, pB])
        return LineString([(pA.x, pA.y), (pB.x, pB.y)])

    @staticmethod
    def __keepTheLargestSolidAngle(sensor, masks):
        sensor = GeomLib3D.centroid(sensor)
        if isinstance(masks, (Point, LineString)):
            _, remotePoint = nearest_points(sensor, masks)
            # if equals(sensor, remotePoint):
            #     return LineString([])
            # return LineString([sensor, remotePoint])
            return RayCasting4Lib.__build_LineString(sensor, remotePoint)

        maxSolidAngle, maxMask = -float("inf"), None
        for mask in masks.geoms:
            h, w = mask.coords[0][2], sensor.distance(mask)
            if (0 == w):
                return LineString([])
            currSolidAngle = h / w
            if (currSolidAngle > maxSolidAngle):
                maxSolidAngle, maxMask = currSolidAngle, mask
        _, remotePoint = nearest_points(sensor, maxMask)

        # if equals(sensor, remotePoint):
        #     return LineString([])
        # return LineString([sensor, remotePoint])
        return RayCasting4Lib.__build_LineString(sensor, remotePoint)

    @staticmethod
    def __from25DRaysToRayLengths(nRays, rayLength, ray_ids, mls):
        _raylens = asarray(
            [ray.length for ray in GeomLib.toListOfLineStrings(mls)])

        if (nRays == len(_raylens)):
            return _raylens

        # raylens = asarray([rayLength for _ in range(nRays)])
        raylens = zeros(nRays)
        for i, ray_id in enumerate(ray_ids):
            raylens[ray_id] = _raylens[i]
        return raylens

    @staticmethod
    def __from25DRaysToRayAlts(nRays, ray_ids, mls):
        _rayalts = asarray(
            [ray.coords[-1][2] for ray in GeomLib.toListOfLineStrings(mls)])

        if (nRays == len(_rayalts)):
            return _rayalts

        # The altitude of the building adjacent to the viewpoint should be assigned here.
        rayalts = zeros(nRays)
        for i, ray_id in enumerate(ray_ids):
            rayalts[ray_id] = _rayalts[i]
        return rayalts

    @staticmethod
    def __from25DRaysToRayDeltaAlts(nRays, ray_ids, mls):
        _rayDeltaAlts = asarray(
            [max(ray.coords[-1][2]-ray.coords[0][2], 0) for ray in GeomLib.toListOfLineStrings(mls)])

        if (nRays == len(_rayDeltaAlts)):
            return _rayDeltaAlts

        # The altitude of the building adjacent to the viewpoint should be assigned here.
        rayDeltaAlts = zeros(nRays)
        for i, ray_id in enumerate(ray_ids):
            rayDeltaAlts[ray_id] = _rayDeltaAlts[i]
        return rayDeltaAlts

    @staticmethod
    def multipleRayCast25D(buildings, rays, nRays, rayLength, elevationFieldName, withIndices, h0=0.0):
        if not GeoDataFrameLib.shareTheSameCrs(buildings, rays):
            raise Exception(
                "Illegal argument: buildings and rays must share shames CRS!")
        rays.geometry = rays.geometry.apply(lambda geom: GeomLib.forceZCoordinateToZ0(
            geom, z0=geom.coords[0][2] if geom.has_z else h0))

        smapRaysField = overlay(rays, buildings.geometry.to_frame(
        ), how="intersection", keep_geom_type=True)
        smapRaysField = smapRaysField.dissolve(by=["__VPT_ID__", "__RAY_ID__"], as_index=False, aggfunc={
                                               "gid": "first", "viewpoint": "first"})
        smapRaysField.geometry = smapRaysField.apply(lambda row: RayCasting4Lib.__keepTheLargestSolidAngle(
            row.viewpoint, row.geometry), axis=1)
        # smapRaysField.to_csv("/tmp/a.csv", index=False, sep=";")

        left = rays.set_index(["__VPT_ID__", "__RAY_ID__"], drop=False)
        left.drop(columns=["__VPT_ID__"], inplace=True)
        right = smapRaysField.set_index(
            ["__VPT_ID__", "__RAY_ID__"]).geometry.to_frame()
        smapRaysField = left.merge(
            right, how="outer", left_index=True, right_index=True)

        smapRaysField["geometry"] = smapRaysField.apply(
            lambda row: row.geometry_x if row.geometry_y is None else row.geometry_y, axis=1)
        smapRaysField.drop(columns=["geometry_x", "geometry_y"], inplace=True)
        smapRaysField = GeoDataFrame(smapRaysField, crs=buildings.crs)
        smapRaysField = smapRaysField.loc[smapRaysField[~smapRaysField.geometry.apply(
            lambda geom: geom.is_empty)].index, :]
        smapRaysField = smapRaysField.dissolve(
            by="__VPT_ID__", as_index=False, aggfunc=list)
        smapRaysField.drop(columns=["__VPT_ID__"], inplace=True)
        for fieldname in smapRaysField.columns:
            if fieldname not in ["geometry", "__RAY_ID__"]:
                smapRaysField[fieldname] = smapRaysField[fieldname].apply(
                    lambda t: t[0])

        if (0 < len(smapRaysField)):
            smapRaysField["__RAY_LEN__"] = smapRaysField.apply(
                lambda row: RayCasting4Lib.__from25DRaysToRayLengths(nRays, rayLength, row.__RAY_ID__, row.geometry), axis=1)
            smapRaysField["__RAY_ALT__"] = smapRaysField.apply(
                lambda row: RayCasting4Lib.__from25DRaysToRayAlts(nRays, row.__RAY_ID__, row.geometry), axis=1)
            smapRaysField["__RAY_DELTA_ALT__"] = smapRaysField.apply(
                lambda row: RayCasting4Lib.__from25DRaysToRayDeltaAlts(nRays, row.__RAY_ID__, row.geometry), axis=1)

            if withIndices:
                smapRaysField = RayCasting4Lib.addIndicesToSkyMapRaysField25D(
                    nRays, smapRaysField)

        return smapRaysField

    @staticmethod
    def addIndicesToSkyMapRaysField25D(nRays, smapRaysField, precision=1.0, base=exp(1)):
        smapRaysField["w_mean"] = smapRaysField.__RAY_LEN__.apply(
            lambda raylens: float(raylens.mean()))
        smapRaysField["w_std"] = smapRaysField.__RAY_LEN__.apply(
            lambda raylens: float(raylens.std()))
        smapRaysField["h_mean"] = smapRaysField.__RAY_ALT__.apply(
            lambda heights: float(heights.mean()))

        smapRaysField["h_over_w"] = smapRaysField.apply(
            lambda row: row.__RAY_DELTA_ALT__.mean() / (2 * row.w_mean), axis=1)
        smapRaysField["svf"] = smapRaysField.apply(
            lambda row: SVFLib.svf2018(row.__RAY_DELTA_ALT__, row.__RAY_LEN__), axis=1)

        return smapRaysField



'''
Created on 8 juin 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import warnings

from geopandas import GeoDataFrame
from pandas import read_csv
from shapely.geometry import LineString, Point
from t4gpd.commons.WarnUtils import WarnUtils
from t4gpd.io.AbstractReader import AbstractReader


class UClimTrackWaypointsReader(AbstractReader):
    '''
    classdocs
    '''

    def __init__(self, inputFile, icrs="epsg:4326", ocrs="epsg:2154"):
        '''
        Constructor
        '''
        warnings.formatwarning = WarnUtils.format_Warning_alt
        self.inputFile = inputFile
        self.icrs = icrs
        self.ocrs = ocrs

    def __getVersion(self):
        return 1

    def __postpcs(self, df):
        df["geometry"] = df.apply(lambda row: Point((row.long, row.lat)), axis=1)
        gdf = GeoDataFrame(df, crs=self.icrs).to_crs(self.ocrs)
        assert gdf.ptid.is_monotonic_increasing, "ptid must be monotonic increasing!"

        track_geom = LineString(gdf.geometry.apply(lambda g: g.coords[0:2][0]))
        tracks = GeoDataFrame([{"length": track_geom.length, "geometry": track_geom}], crs=self.ocrs)

        waypoints = gdf.loc[gdf.waypoint == 1 , ["ptid", "geometry"]]
        waypoints.reset_index(drop=True, inplace=True)
        waypoints["curv_absc"] = waypoints.geometry.apply(lambda g: track_geom.project(g, normalized=False))
        waypoints["sect_len"] = waypoints.curv_absc.diff(periods=-1).abs()
        waypoints.rename(columns={"ptid": "id"}, inplace=True)

        return tracks, waypoints

    def __readV1(self):
        df = read_csv(UClimTrackWaypointsReader.opener(self.inputFile), sep=";")
        return self.__postpcs(df)

    def run(self):
        version = self.__getVersion()
        if (1 == version):
            warnings.warn("UClimTrackWaypointsReader (v1)")
            return self.__readV1()
        raise Exception("Unreachable instruction!")

"""
import matplotlib.pyplot as plt

ifile = "/home/tleduc/prj/uclim/data/nantes_commerce_feydeau/nantes_commerce_feydeau_track1.csv"
tracks, waypoints = UClimTrackWaypointsReader(ifile).run()

gdf.to_csv("/tmp/a.csv", index=False, sep=";")
print(gdf.head(3))

ax=waypoints.plot(); tracks.plot(ax=ax); plt.show()
"""



'''
Created on 12 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import isnan
from pandas.core.frame import DataFrame
from t4gpd.comfort.algo.SETLib import SETLib
from t4gpd.comfort.indices.AbstractThermalComfortIndice import AbstractThermalComfortIndice
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class SET(AbstractThermalComfortIndice):
    '''
    classdocs
    '''

    def __init__(self, sensorsGdf, AirTC='AirTC_Avg', RH='RH_Avg', WS_ms='WS_ms_Avg',
                 T_mrt='T_mrt'):
        '''
        Constructor

        AirTC: air temperature [C]
        RH: relative humidity [%]
        Ws_ms: wind speed recorded at pedestrian level (at height 1.1 m) [m.s-1]
        T_mrt: Mean radiant temperature [C]

        SET: Standard Effective Temperature
        '''
        if not isinstance(sensorsGdf, DataFrame):
            raise IllegalArgumentTypeException(sensorsGdf, 'DataFrame')

        for fieldname in (AirTC, RH, WS_ms, T_mrt):
            if fieldname not in sensorsGdf:
                raise Exception('%s is not a relevant field name!' % fieldname)

        self.AirTC = AirTC
        self.RH = RH
        self.WS_ms = WS_ms
        self.T_mrt = T_mrt

    @staticmethod
    def thermalPerceptionRanges():
        # Excerpt from https://doi.org/10.1016/j.wace.2018.01.004
        return {
            'Moderate cold': { 'min':-float('inf'), 'max': 17, 'color': '#91c1e1' },
            'Comfortable': { 'min': 17, 'max': 30, 'color': '#ffffff' },
            'Moderate heat': { 'min': 30, 'max': 34, 'color': '#fffa00' },
            'Strong heat': { 'min': 34, 'max': 37, 'color': '#ff7900' },
            'Extreme heat': { 'min': 37, 'max': float('inf'), 'color': '#ff0000' }
            }

    def runWithArgs(self, row):
        AirTC = row[self.AirTC]
        RH = row[self.RH]
        WS_ms = row[self.WS_ms]
        T_mrt = row[self.T_mrt]

        SET = None
        if not (isnan(AirTC) or isnan(RH) or isnan(WS_ms) or isnan(T_mrt)):
            # SET: Standard Effective Temperature
            SET = SETLib.assess_set(AirTC, RH, WS_ms, T_mrt)

        return { 'SET': SET }



'''
Created on 8 juin 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from pandas import DataFrame, Interval, IntervalIndex
from t4gpd.commons.GeoDataFrameLib import GeoDataFrameLib
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class SnapUclimOnTrackUsingWaypoints(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, dfUclim, dfMob, tracks, waypoints):
        '''
        Constructor
        '''
        if not isinstance(dfUclim, DataFrame):
            raise IllegalArgumentTypeException(dfUclim, "DataFrame")
        for fieldname in ["wp1", "wp2", "timestamps"]:
            if fieldname not in dfUclim:
                raise Exception(f"{fieldname} is not a relevant field name!")
        assert dfUclim.wp1.is_monotonic_increasing, "dfImu.wp1 must be increasing!"
        assert dfUclim.wp1.is_unique, "dfImu.wp1 must be strictly increasing!"
        assert dfUclim.wp2.is_monotonic_increasing, "dfImu.wp2 must be increasing!"
        assert dfUclim.wp2.is_unique, "dfImu.wp2 must be strictly increasing!"
        self.dfUclim = dfUclim

        if not isinstance(dfMob, DataFrame):
            raise IllegalArgumentTypeException(dfMob, "DataFrame")
        self.dfMob = dfMob

        if not isinstance(tracks, GeoDataFrame):
            raise IllegalArgumentTypeException(tracks, "GeoDataFrame")
        self.track_geom = tracks.geometry.squeeze()

        if not isinstance(waypoints, GeoDataFrame):
            raise IllegalArgumentTypeException(waypoints, "GeoDataFrame")
        if "id" not in waypoints:
            raise Exception(f"'id' is not a relevant field name!")
        self.waypoints = waypoints

        assert GeoDataFrameLib.shareTheSameCrs(
            tracks, waypoints), "tracks and waypoints must share the same crs!"
        self.crs = tracks.crs

    def _buildIntervalIndex(self):
        intervals, rows = [], []
        for _, row in self.dfUclim.iterrows():
            wp1, dts = row.wp1, row.timestamps
            for idx2 in range(0, len(dts), 2):
                intervals.append(
                    Interval(left=dts[idx2], right=dts[idx2 + 1], closed="right"))
                rows.append(wp1)
        return DataFrame(data=rows, columns=["wp1"], index=IntervalIndex(intervals))

    def _getWps(self, intervals, dt):
        idx = intervals.index.get_loc(dt)
        row = intervals.iloc[idx]
        return row.wp1

    def run(self):
        intervals = self._buildIntervalIndex()

        def _isInIntervalIndex(intervals, dt): return any(
            intervals.index.contains(dt))

        # SELECTION FROM THE SET OF MEASUREMENTS OF THOSE THAT FALL WITHIN THE INTERVALS
        measures1 = self.dfMob[self.dfMob.timestamp.apply(
            lambda dt: _isInIntervalIndex(intervals, dt))]
        measures1 = measures1[["timestamp"]]
        measures1["wp1"] = measures1.timestamp.apply(
            lambda dt: self._getWps(intervals, dt))
        measures1["inc"] = 1

        measures2 = measures1.groupby(by="wp1").\
            inc.expanding().sum().reset_index(level=[0])
        measures2.inc = measures2.inc.astype(int)
        measures2 = measures2.merge(measures2.groupby(by="wp1").transform("max"),
                                    how="inner", left_index=True, right_index=True)

        # ATTRIBUTE JOIN BETWEEN THE GIVEN TRACK AND ALL WAYPOINTS
        wp = self.dfUclim[["warning", "actual_elapsed_time"]].merge(
            self.waypoints[["id", "curv_absc", "sect_len"]],
            how="left", left_index=True, right_index=True)
        wp["sect_speed"] = wp.apply(
            lambda row: row.sect_len/row.actual_elapsed_time, axis=1)

        # ATTRIBUTE JOIN BETWEEN SELECTED MEASUREMENTS AND ALL WAYPOINTS
        measures2 = measures2.merge(wp[["curv_absc", "sect_len", "actual_elapsed_time", "sect_speed", "warning"]],
                                    how="inner", left_on="wp1", right_index=True)
        measures2.rename(columns={"inc_x": "inc", "inc_y": "total", "curv_absc": "curv_absc0"},
                         inplace=True)

        # DISTRIBUTION OF MEASUREMENTS ON THE POLYLINE ACCORDING TO THEIR CURVILINEAR ABSCISSA VALUE
        measures2["track"] = 1
        measures2["curv_absc"] = measures2.apply(
            lambda row: row.curv_absc0 + (row.inc * row.sect_len) / row.total,
            axis=1)
        measures2["geometry"] = measures2.curv_absc.apply(
            lambda x: self.track_geom.interpolate(x, normalized=False))

        # ADD COLUMNS "ptid" AND "timestamp" TO THE MEASUREMENTS DATAFRAME
        measures3 = measures2[["track", "inc", "curv_absc", "sect_len",
                               "actual_elapsed_time", "sect_speed", "warning", "geometry"]].merge(
            measures1[["timestamp"]], how="inner", left_index=True, right_index=True)
        measures3 = measures3.merge(
            self.waypoints[["id", "curv_absc"]], how="left", on="curv_absc")
        measures3.rename(columns={"id": "ptid"}, inplace=True)

        measures3.at[0, "ptid"] = 1
        measures3 = GeoDataFrame(measures3, crs=self.crs)

        assert measures3.curv_absc.is_monotonic_increasing, "measures3.curv_absc must be increasing!"
        assert measures3.curv_absc.is_unique, "measures3.curv_absc must be strictly increasing!"

        # return intervals, measures1, measures2, measures3
        return measures3


"""
from t4gpd.picoclim.CampbellSciReader import CampbellSciReader
from t4gpd.picoclim.UClimGuidingReader import UClimGuidingReader
from t4gpd.picoclim.UClimTrackWaypointsReader import UClimTrackWaypointsReader

dir1 = "/home/tleduc/prj/uclim/flask/static/uploads/1686227346623113409"
dir2 = "/home/tleduc/prj/uclim/data/nantes_commerce_feydeau"
dfUclim = UClimGuidingReader(f"{dir1}/uclimLOC_nantes_commerce_feydeau_20230607T1228.txt").run()
tracks, waypoints = UClimTrackWaypointsReader(f"{dir2}/nantes_commerce_feydeau_track1.csv").run()
dfMob = CampbellSciReader(f"{dir1}/CR1000XSeries_TwoSec.dat").run() 
# r0, r1, r2, r3 = SnapUclimOnTrackUsingWaypoints(dfUclim, dfMob, tracks, waypoints).run()
r3 = SnapUclimOnTrackUsingWaypoints(dfUclim, dfMob, tracks, waypoints).run()

import matplotlib.pyplot as plt
fig, ax = plt.subplots(figsize=(1.5 * 8.26, 1.5 * 8.26))
waypoints.plot(ax=ax, marker="P", color="red")
r3.plot(ax=ax, marker="+", color="black")
tracks.plot(ax=ax, color="red", linewidth=0.3)
ax.axis("off")
fig.tight_layout()
plt.show()

waypoints.to_file("/tmp/1.gpkg", layer="waypoints")
tracks.to_file("/tmp/1.gpkg", layer="tracks")
r3.to_file("/tmp/1.gpkg", layer="measures")
"""



'''
Created on 28 mar. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame, sjoin
from numpy import arange, array, delete
from numpy.random import shuffle
from shapely.geometry import Polygon
from shapely.geometry.multipolygon import MultiPolygon
from shapely.ops import unary_union
from shapely.wkt import loads
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.morph.STFacadesAnalysis import STFacadesAnalysis

from t4gpd.commons.GridFaceLib import GridFaceLib


class STDeploymentOfPVPanels3(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, buildings, idFieldName='ID', coverRate=1.0, ribbonWidth=3.2,
                 azimuthRange=(270 - 45, 270 + 45), sizeOfPVPanel=(1.6, 1.0)):
        '''
        Constructor
        '''
        if not isinstance(buildings, GeoDataFrame):
            raise IllegalArgumentTypeException(buildings, 'buildings GeoDataFrame')
        self.buildings = buildings

        if idFieldName not in buildings:
            raise Exception('%s is not a relevant field name!' % (idFieldName))
        self.idFieldName = idFieldName

        self.coverRate = coverRate
        self.ribbonWidth = ribbonWidth
        self.azim0, self.azim1 = azimuthRange
        self.sizeOfPVPanel = sizeOfPVPanel

    @staticmethod
    def __nModules(ribbonLength, ribbonWidth, sizeOfPVPanel):
        d1, d2 = sizeOfPVPanel
        return max(
            (ribbonLength // d1) * (ribbonWidth // d2),
            (ribbonLength // d2) * (ribbonWidth // d1)
            )

    @staticmethod
    def __buildPvArea(edge, normal, ribbonWidth):
        p0, p1 = edge.coords
        n0, n1 = loads(normal).coords
        n = (n1[0] - n0[0], n1[1] - n0[1])
        return Polygon([
            p0, (p0[0] - ribbonWidth * n[0], p0[1] - ribbonWidth * n[1]),
            (p1[0] - ribbonWidth * n[0], p1[1] - ribbonWidth * n[1]), p1
            ])

    def __onRoofRibbon(self):
        # ------
        # REMOVE SHARED WALLS
        uob = unary_union(self.buildings.geometry).buffer(1e-4)
        uob = GeoDataFrame([{'geometry': uob}], crs=self.buildings.crs)
        uob = uob.explode(ignore_index=True)

        # ------
        # SELECT WELL ORIENTED AND BIG ENOUGH FACADES
        facades = STFacadesAnalysis(uob, gidFieldname=None, elevationFieldname=None,
                                    exteriorOnly=False).run()
        facades = facades[(facades.azimuth >= self.azim0) & (facades.azimuth <= self.azim1) & 
                          (facades.length > (0.5 + min(self.sizeOfPVPanel)))]

        facades.drop(columns=['buildingID', 'height', 'surface'], inplace=True)

        # ------
        # SPATIAL JOIN BETWEEN FACADES AND BUILDINGS
        _facades = facades.copy()
        _facades.geometry = _facades.geometry.apply(lambda g: g.centroid.buffer(1e-2))
        _facades = sjoin(_facades[['gid', 'geometry']],
                         self.buildings[[self.idFieldName, 'geometry']],
                         how='left', predicate='intersects')
        _facades.drop(columns=['index_right', 'geometry'], inplace=True)

        # ------
        # ATTRIBUTE JOIN BETWEEN FACADES AND _FACADES
        pvRibbon = facades.merge(_facades, on='gid')

        # ------
        # NUMBER OF PHOTOVOLTAIC MODULES PER FACADES
        pvRibbon['n_modules'] = pvRibbon.length.apply(
            lambda l: self.__nModules(l, self.ribbonWidth, self.sizeOfPVPanel))
        pvRibbon.n_modules = self.coverRate * pvRibbon.n_modules  # apply the coverage rate
        pvRibbon = pvRibbon.loc[ pvRibbon[pvRibbon.n_modules > 1].index ]  # Remove facades without modules
        pvRibbon.n_modules = pvRibbon.n_modules.astype(int)

        # ------
        # FROM FACADES TO PV AREA
        pvRibbon['__TMP__'] = list(zip(pvRibbon.geometry, pvRibbon.normal))
        pvRibbon.geometry = pvRibbon.__TMP__.apply(lambda t: self.__buildPvArea(*t, self.ribbonWidth))
        pvRibbon.drop(columns=['__TMP__'], inplace=True)

        return pvRibbon

    def __onFlatRoof(self):
        rows = []
        for _, row in self.buildings.iterrows():
            geom = row.geometry
            result = GridFaceLib.grid(geom, dx=self.sizeOfPVPanel[0],
                                      dy=self.sizeOfPVPanel[1], intoPoint=False)
            geom, n_modules = result['geometry'], result['n_cells']

            nItems = len(geom.geoms)
            nItemsToDel = int((1.0 - self.coverRate) * nItems)
            idxToDel = arange(nItems)
            shuffle(idxToDel)
            idxToDel = sorted(idxToDel[0:nItemsToDel], reverse=True)
            geom = delete(array(geom.geoms), idxToDel)
            n_modules = len(geom)
            geom = MultiPolygon(geom.tolist())

            rows.append(self.updateOrAppend(row, { 'geometry': geom, 'n_modules': n_modules }))

        return GeoDataFrame(rows, crs=self.buildings.crs)

    def run(self):
        if (self.ribbonWidth is None) or (self.azim0 is None) or (self.azim1 is None):
            return self.__onFlatRoof()
        return self.__onRoofRibbon()



'''
Created on 22 sep. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import matplotlib.pyplot as plt
from datetime import datetime, timedelta, timezone
from geopandas import GeoDataFrame
from locale import LC_ALL, setlocale
from numpy import sqrt
from shapely import intersection_all, Point, union_all
from shapely.affinity import rotate, translate
from shapely.geometry import CAP_STYLE
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.LatLonLib import LatLonLib
from t4gpd.commons.sun.SunLib import SunLib
from t4gpd.sun.STHardShadow import STHardShadow


class WilliamAtkinson(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, rotation=0, position=LatLonLib.NANTES,
                 model="pysolar"):
        '''
        Constructor
        '''
        if ("epsg:4326" != position.crs):
            raise IllegalArgumentTypeException(
                position, "Must be a WGS84 GeoDataFrame")
        self.lat, _ = LatLonLib.fromGeoDataFrameToLatLon(position)
        self.sunModel = SunLib(position, model=model)
        # gdf = position.to_crs(position.estimate_utm_crs(datum_name="WGS 84")) # metric CRS
        # self.x, self.y = gdf.geometry.squeeze().x, gdf.geometry.squeeze().y
        self.rotation = rotation

    def mockup(self, angle):
        geom = Point((0, 0)).buffer(0.5, cap_style=CAP_STYLE.square)
        geom = rotate(geom, angle=angle, origin="center", use_radians=False)
        # geom = translate(geom, xoff=self.x, yoff=self.y)
        return GeoDataFrame([{"geometry": geom, "HAUTEUR": 1.0}])

    def shadows(self, days, angles, delta=30, ofile=None):
        nrows, ncols = len(days), len(angles)
        fig, axes = plt.subplots(nrows=nrows, ncols=ncols, figsize=(
            ncols * 8.26, 8.26), squeeze=False)

        for nr, day in enumerate(days):
            sunrise = self.sunModel.getSunrise(day)
            sunrise += timedelta(minutes=delta)
            sunset = self.sunModel.getSunset(day)
            sunset -= timedelta(minutes=delta)
            # dts = arange(sunrise, sunset, timedelta(minutes=10))
            dts = [(sunrise, sunset, timedelta(minutes=10))]

            for nc, angle in enumerate(angles):
                ax = axes[nr, nc]
                building = self.mockup(angle)
                setlocale(LC_ALL, "en_US.utf8")
                shadows = STHardShadow(building, dts,
                                       occludersElevationFieldname="HAUTEUR",
                                       altitudeOfShadowPlane=0, aggregate=True).run()
                area1 = intersection_all(shadows.geometry).area
                area2 = union_all(shadows.geometry).area

                ax.set_title(
                    f"{sunrise.strftime('%b %d: %H:%M')} - {sunset.strftime('%H:%M')}, {self.lat:+.1f}°N, {area1:.1f} m$^2$, {area2:.1f} m$^2$")
                shadows.boundary.plot(ax=ax, color="grey", linewidth=0.5)
                building.plot(ax=ax, color="black")

        if ofile is None:
            fig.tight_layout()
            plt.show()
        else:
            plt.savefig(ofile, bbox_inches="tight")
        plt.close(fig)


"""
WilliamAtkinson().shadows(days=[datetime(2023, 12, 21), datetime(
    2023, 3, 21), datetime(2023, 6, 21)], angles=[0, 45])
"""



'''
Created on 28 avr. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import timedelta, timezone
from io import StringIO
from os.path import basename
import re
import warnings
from zoneinfo import ZoneInfo

from geopandas import GeoDataFrame
from pandas import DataFrame, to_datetime
from shapely import LineString, Point
from t4gpd.commons.WarnUtils import WarnUtils
from t4gpd.io.AbstractReader import AbstractReader


class UClimGuidingReader(AbstractReader):
    '''
    classdocs
    '''

    def __init__(self, inputFile, generateTrackAndWaypoints=False, crs="EPSG:4326"):
        '''
        Constructor
        '''
        warnings.formatwarning = WarnUtils.format_Warning_alt
        self.inputFile = inputFile
        self.generateTrackAndWaypoints = generateTrackAndWaypoints
        self.crs = crs
        self.dfUclim = None

    def __getVersion(self):
        RE_VERSION = re.compile(r"^#\s*CODING\s+VERSION\s+(\d+)$")
        with UClimGuidingReader.opener(self.inputFile) as f:
            for line in f:
                line = line.strip().upper()
                if RE_VERSION.match(line):
                    return int(RE_VERSION.search(line).group(1))
        return 1

    def __check(self):
        df = self.getDfUclim()

        assert df.wp1.is_monotonic_increasing, "The wp1 column is not increasing"
        assert df.wp1.is_unique, "The wp1 column is not strictly increasing"

        assert df.wp2.is_monotonic_increasing, "The wp2 column is not increasing"
        assert df.wp2.is_unique, "The wp2 column is not strictly increasing"

        assert df.timestamps.explode().is_monotonic_increasing, "Timestamps are not ascending"
        assert df.timestamps.explode().is_unique, "Timestamps are not strictly ascending"

    def __generateTrackAndWaypoints(self, points, waypoints):
        waypoints = [{"id": 100 + i, "geometry": points[waypoints[i]]} for i in sorted(waypoints.keys())]
        waypoints = GeoDataFrame(waypoints, crs=self.crs)

        points = [points[i] for i in sorted(points.keys())]
        track = LineString([(p.x, p.y) for p in points])
        track = GeoDataFrame([{"geometry": track}], crs=self.crs)

        return track, waypoints

    @staticmethod
    def _timestamp2datetime(ts, tzinfo=ZoneInfo("Europe/Paris")):
        return to_datetime(int(ts), unit="ms").\
            tz_localize(timezone.utc).tz_convert(tzinfo)

    def __readV1(self):
        RE_ROUTE_ID = re.compile(r"^#\s*ROUTE_ID\s+(.*)$")
        RE_COMMENT = re.compile(r"^#")
        RE_CANCELING = re.compile(r"^.*CANCELING")
        RE_PREPARING = re.compile(r"^PREPARING\s+MOVE\s+FROM\s+WP_(\d+)\s+TO\s+WP_(\d+)$")
        RE_STARTING = re.compile(r"^STARTING AT\s+\d{13}\s+(STOPPING\s+AT\s+\d{13}\s+RESUMING\s+AT\s+\d{13}\s+)*ARRIVING\s+AT\s+\d{13}$")
        RE_WP = re.compile(r"^WP_(\d+)\s+PT_(\d+)$")
        RE_PT = re.compile(r"^PT_(\d+)\s+(\d+[\.]?\d*)\s+(\d+[\.]?\d*)$")
        RE_TIMESTAMP = re.compile(r"\d{13}")

        tzinfo = ZoneInfo("Europe/Paris")
        waypoints, points, df = {}, {}, {}
        with UClimGuidingReader.opener(self.inputFile) as f:
            for nline, line in enumerate(f, start=1):
                line = line.strip().upper()

                if RE_ROUTE_ID.match(line):
                    route_id = RE_ROUTE_ID.search(line).group(1)

                elif RE_COMMENT.match(line) or RE_CANCELING.match(line) or (0 == len(line)):
                    # There is nothing to do
                    pass

                elif RE_PREPARING.match(line):
                    wp1, wp2 = RE_PREPARING.search(line).groups()
                    wp1, wp2 = int(wp1), int(wp2) 
                    assert wp2 == 1 + wp1

                elif RE_STARTING.match(line):
                    dts = RE_TIMESTAMP.findall(line)
                    dts = [self._timestamp2datetime(dt, tzinfo) for dt in dts]
                    # dts = [dt.strftime("%X") for dt in dts]
                    df[wp1] = dts

                elif RE_WP.match(line):
                    wp, pt = RE_WP.search(line).groups()
                    waypoints[int(wp)] = int(pt)

                elif RE_PT.match(line):
                    pt, x, y = RE_PT.search(line).groups()
                    points[int(pt)] = Point((x, y))

                else:
                    raise Exception(f"Syntax error line {nline}")

        rows = []
        for wp1 in sorted(df.keys()):
            rows.append({"route_id": route_id, "wp1": wp1, "wp2": wp1 + 1, "timestamps": df[wp1]})
        df = DataFrame(rows)

        if self.generateTrackAndWaypoints:
            track, waypoints = self.__generateTrackAndWaypoints(points, waypoints)
            return df, track, waypoints
        return df

    def __readV2(self):
        RE_ROUTE_ID = re.compile(r"^#\s*ROUTE_ID\s+(.*)$", flags=re.IGNORECASE)
        RE_TIMEZONE = re.compile(r"^#\s*TIMEZONE\s+(.*)$", flags=re.IGNORECASE)
        RE_COMMENT = re.compile(r"^#")
        RE_CANCELING = re.compile(r"^.*CANCELING", flags=re.IGNORECASE)
        RE_PREPARING = re.compile(r"^MOVING\s+FROM\s+WP_(\d+)\s+TO\s+WP_(\d+)$", flags=re.IGNORECASE)
        RE_STARTING = re.compile(r"^STARTING AT\s+\d{13}\s+(PAUSING\s+AT\s+\d{13}\s+RESUMING\s+AT\s+\d{13}\s+)*ARRIVING\s+AT\s+\d{13}$", flags=re.IGNORECASE)
        RE_WARNING = re.compile(r"^WARNING\s*:(.*)$", flags=re.IGNORECASE)
        RE_WP = re.compile(r"^WP_(\d+)\s+PT_(\d+)$", flags=re.IGNORECASE)
        RE_PT = re.compile(r"^PT_(\d+)\s+(\d+[\.]?\d*)\s+(\d+[\.]?\d*)$", flags=re.IGNORECASE)
        RE_TIMESTAMP = re.compile(r"\d{13}")

        tzinfo = None
        waypoints, points, df = {}, {}, {}
        with UClimGuidingReader.opener(self.inputFile) as f:
            for nline, line in enumerate(f, start=1):
                line = line.strip()

                if RE_ROUTE_ID.match(line):
                    route_id = RE_ROUTE_ID.search(line).group(1)

                elif RE_TIMEZONE.match(line):
                    tzinfo = ZoneInfo(RE_TIMEZONE.search(line).group(1))

                elif RE_COMMENT.match(line) or RE_CANCELING.match(line) or (0 == len(line)):
                    # There is nothing to do
                    pass

                elif RE_PREPARING.match(line):
                    warning_msg = None
                    wp1, wp2 = RE_PREPARING.search(line).groups()
                    wp1, wp2 = int(wp1), int(wp2) 
                    assert wp2 == 1 + wp1

                elif RE_STARTING.match(line):
                    dts = RE_TIMESTAMP.findall(line)
                    dts = [self._timestamp2datetime(dt, tzinfo) for dt in dts]
                    # dts = [dt.strftime("%X") for dt in dts]
                    df[wp1] = {"timestamps": dts}

                elif RE_WARNING.match(line):
                    warning_msg = RE_WARNING.search(line).group(1).strip()
                    df[wp1]["warning"] = warning_msg

                elif RE_WP.match(line):
                    wp, pt = RE_WP.search(line).groups()
                    waypoints[int(wp)] = int(pt)

                elif RE_PT.match(line):
                    pt, x, y = RE_PT.search(line).groups()
                    points[int(pt)] = Point((x, y))

                else:
                    raise Exception(f"Syntax error line {nline}\n\t{line}")

        rows = []
        for wp1 in sorted(df.keys()):
            _timestamps = df[wp1]["timestamps"]
            _warning = df[wp1]["warning"] if ("warning" in df[wp1]) else None
            rows.append({"route_id": route_id, "wp1": wp1, "wp2": wp1 + 1,
                         "timestamps": _timestamps, "warning": _warning})
        df = DataFrame(rows)

        if self.generateTrackAndWaypoints:
            track, waypoints = self.__generateTrackAndWaypoints(points, waypoints)
            return df, track, waypoints
        return df

    def __readV3(self):
        RE_ROUTE_ID = re.compile(r"^#\s*ROUTE_ID\s+(.*)$", flags=re.IGNORECASE)
        RE_TRACK_ID = re.compile(r"^#\s*TRACK\s+(.*)$", flags=re.IGNORECASE)
        RE_TIMEZONE = re.compile(r"^#\s*TIMEZONE\s+(.*)$", flags=re.IGNORECASE)
        RE_COMMENT = re.compile(r"^#")
        RE_CANCELING = re.compile(r"^.*CANCELING", flags=re.IGNORECASE)
        RE_PREPARING = re.compile(r"^MOVING\s+FROM\s+WP_(\d+)\s+TO\s+WP_(\d+)$", flags=re.IGNORECASE)
        RE_STARTING = re.compile(r"^STARTING AT\s+\d{13}\s+(PAUSING\s+AT\s+\d{13}\s+RESUMING\s+AT\s+\d{13}\s+)*ARRIVING\s+AT\s+\d{13}$", flags=re.IGNORECASE)
        RE_WARNING = re.compile(r"^WARNING\s*:(.*)$", flags=re.IGNORECASE)
        RE_TIMESTAMP = re.compile(r"\d{13}")

        tzinfo = None
        df = {}
        with UClimGuidingReader.opener(self.inputFile) as f:
            for nline, line in enumerate(f, start=1):
                line = line.strip()

                if RE_ROUTE_ID.match(line):
                    route_id = RE_ROUTE_ID.search(line).group(1)

                if RE_TRACK_ID.match(line):
                    track_id = RE_TRACK_ID.search(line).group(1)

                elif RE_TIMEZONE.match(line):
                    tzinfo = ZoneInfo(RE_TIMEZONE.search(line).group(1))

                elif RE_COMMENT.match(line) or RE_CANCELING.match(line) or (0 == len(line)):
                    # There is nothing to do
                    pass

                elif RE_PREPARING.match(line):
                    warning_msg = None
                    wp1, wp2 = RE_PREPARING.search(line).groups()
                    wp1, wp2 = int(wp1), int(wp2) 
                    assert wp2 == 1 + wp1

                elif RE_STARTING.match(line):
                    dts = RE_TIMESTAMP.findall(line)
                    dts = [self._timestamp2datetime(dt, tzinfo) for dt in dts]
                    # dts = [dt.strftime("%X") for dt in dts]
                    df[wp1] = {"timestamps": dts}

                elif RE_WARNING.match(line):
                    warning_msg = RE_WARNING.search(line).group(1).strip()
                    df[wp1]["warning"] = warning_msg

                else:
                    raise Exception(f"Syntax error line {nline}\n\t{line}")

        rows = []
        for wp1 in sorted(df.keys()):
            _timestamps = df[wp1]["timestamps"]
            _warning = df[wp1]["warning"] if ("warning" in df[wp1]) else None
            rows.append({"route_id": route_id, "track_id": track_id, "wp1": wp1,
                         "wp2": wp1 + 1, "timestamps": _timestamps,
                         "warning": _warning})
        return DataFrame(rows)

    def __readV4(self):
        RE_PROJECT_ID = re.compile(r"^#\s*PROJECT\s+(.*)$", flags=re.IGNORECASE)
        RE_TRACK_ID = re.compile(r"^#\s*TRACK\s+(.*)$", flags=re.IGNORECASE)
        RE_TIMEZONE = re.compile(r"^#\s*TIMEZONE\s+(.*)$", flags=re.IGNORECASE)
        RE_COMMENT = re.compile(r"^#")
        RE_CANCELING = re.compile(r"^.*CANCELING", flags=re.IGNORECASE)
        RE_PREPARING = re.compile(r"^MOVING\s+FROM\s+WP_(\d+)\s+TO\s+WP_(\d+)$", flags=re.IGNORECASE)
        RE_STARTING = re.compile(r"^STARTING AT\s+\d{13}\s+(PAUSING\s+AT\s+\d{13}\s+RESUMING\s+AT\s+\d{13}\s+)*ARRIVING\s+AT\s+\d{13}$", flags=re.IGNORECASE)
        RE_WARNING = re.compile(r"^WARNING\s*:(.*)$", flags=re.IGNORECASE)
        RE_TIMESTAMP = re.compile(r"\d{13}")

        tzinfo = None
        df = {}
        with UClimGuidingReader.opener(self.inputFile) as f:
            for nline, line in enumerate(f, start=1):
                line = line.strip()

                if RE_PROJECT_ID.match(line):
                    project_id = RE_PROJECT_ID.search(line).group(1)

                if RE_TRACK_ID.match(line):
                    track_id = RE_TRACK_ID.search(line).group(1)

                elif RE_TIMEZONE.match(line):
                    tzinfo = ZoneInfo(RE_TIMEZONE.search(line).group(1))

                elif RE_COMMENT.match(line) or RE_CANCELING.match(line) or (0 == len(line)):
                    # There is nothing to do
                    pass

                elif RE_PREPARING.match(line):
                    warning_msg = None
                    wp1, wp2 = RE_PREPARING.search(line).groups()
                    wp1, wp2 = int(wp1), int(wp2) 
                    assert wp2 == 1 + wp1

                elif RE_STARTING.match(line):
                    dts = RE_TIMESTAMP.findall(line)
                    dts = [self._timestamp2datetime(dt, tzinfo) for dt in dts]
                    # dts = [dt.strftime("%X") for dt in dts]
                    df[wp1] = {"timestamps": dts}

                elif RE_WARNING.match(line):
                    warning_msg = RE_WARNING.search(line).group(1).strip()
                    df[wp1]["warning"] = warning_msg

                else:
                    raise Exception(f"Syntax error line {nline}\n\t{line}")

        rows = []
        for wp1 in sorted(df.keys()):
            _timestamps = df[wp1]["timestamps"]
            _warning = df[wp1]["warning"] if ("warning" in df[wp1]) else None
            rows.append({"project_id": project_id, "track_id": track_id, "wp1": wp1,
                         "wp2": wp1 + 1, "timestamps": _timestamps,
                         "warning": _warning})
        return DataFrame(rows)

    def getDfUclim(self):
        if self.dfUclim is None:
            version = self.__getVersion()
            if (1 == version):
                self.dfUclim = self.__readV1()
            elif (2 == version):
                self.dfUclim = self.__readV2()
            elif (3 == version):
                self.dfUclim = self.__readV3()
            elif (4 == version):
                self.dfUclim = self.__readV4()
            else:
                raise Exception("Unreachable instruction!")
        return self.dfUclim

    def getPerformance(self):
        df = self.getDfUclim()
        dt0 = df.loc[0, "timestamps"][0]
        dt1 = df.loc[len(df) - 1, "timestamps"][-1]
        deltaTotal = (dt1 - dt0).seconds
        deltaActual = df.actual_elapsed_time.sum()
        return dt0, dt1, deltaTotal, deltaActual, 100 * deltaActual / deltaTotal

    @staticmethod
    def timedeltaBetweenTwoWaypoints(dts):
        assert (0 == len(dts) % 2), "List of an odd number of dates"
        delta = timedelta(0)
        for i in range(0, len(dts), 2):
            delta += dts[i + 1] - dts[i]
        return delta

    def audit(self):
        dt0, dt1, deltaTotal, deltaActual, perf = self.getPerformance()

        identifier = ""
        if not isinstance(self.inputFile, StringIO):
            identifier = f"\n**** {basename(self.inputFile) }"

        warnings.warn("\n" + "*"*70 + f"""{identifier}
****\tFROM: {dt0}
****\tTO: {dt1}
****\tTOTAL PERIOD: {deltaTotal//60} min. {deltaTotal%60} sec.
****\tACTUAL PERIOD: {deltaActual//60} min. {deltaActual%60} sec.
****\tMEASUREMENT PERFORMANCE: {perf:.1f} %
****\tNB WAYPOINTS: {len(self.getDfUclim())}
""" + "*"*70)

    def run(self):
        self.dfUclim = self.getDfUclim()
        self.dfUclim["actual_elapsed_time"] = self.dfUclim.timestamps.apply(
            lambda dts: self.timedeltaBetweenTwoWaypoints(dts).total_seconds())
        self.__check()
        self.audit()
        return self.dfUclim

"""
ifile = "/home/tleduc/prj/uclim/flask/static/uploads/1686227346623113409/uclimLOC_nantes_commerce_feydeau_20230607T1228.txt"
df = UClimGuidingReader(ifile).run()
"""



'''
Created on 22 aug. 2023

@author: tleduc

Copyright 2020-2023 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import rasterio
from geopandas import GeoDataFrame
from numpy import float32
from rasterio import features
from rasterio.io import DatasetReader
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.raster.AbstractRasterGeoProcess import AbstractRasterGeoProcess


class STRasterize(AbstractRasterGeoProcess):
    '''
    classdocs
    '''

    def __init__(self, gdf, raster, attr=None, debug=False):
        '''
        Constructor
        '''
        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, "GeoDataFrame")
        self.gdf = gdf

        if not ((attr is None) or (attr in gdf)):
            raise IllegalArgumentTypeException(
                attr, "must be either None or a valid field name")
        self.attr = attr

        if not isinstance(raster, DatasetReader):
            raise IllegalArgumentTypeException(raster, "DatasetReader")
        self.raster = raster
        self.debug = debug

    def run(self):
        if self.attr is None:
            geoms = self.gdf.geometry.to_list()
            rasterized = features.rasterize(geoms,
                                            out_shape=self.raster.shape,
                                            transform=self.raster.transform,
                                            all_touched=False,
                                            fill=0,  # background value
                                            default_value=1,
                                            # merge_alg=MergeAlg.replace,
                                            dtype=float32)

        else:
            pairs = ((geom, value) for geom, value in zip(
                self.gdf.geometry, self.gdf[self.attr]))

            rasterized = features.rasterize(pairs,
                                            out_shape=self.raster.shape,
                                            transform=self.raster.transform,
                                            all_touched=False,
                                            fill=0,  # background value
                                            # merge_alg=MergeAlg.replace,
                                            dtype=float32)

        out_meta = self.raster.meta.copy()
        out_meta.update({
            "driver": "GTiff",
            "height": self.raster.height,
            "width": self.raster.width,
            "transform": self.raster.transform,
            "crs": self.raster.crs,
            "count": 1,
            "dtype": rasterio.float32
        })

        result = self._write_and_load(rasterized, out_meta, self.debug)
        return result


"""
import matplotlib.pyplot as plt
from t4gpd.demos.GeoDataFrameDemos import GeoDataFrameDemos
from t4gpd.morph.STBBox import STBBox
from t4gpd.raster.RTClip import RTClip

buildings = GeoDataFrameDemos.districtGraslinInNantesBuildings()
roi = STBBox(buildings, buffDist=0).run()
raster1 = rasterio.open(
    "/home/tleduc/data/mnt_asc/mnt_nantes_2005_1m/MNT_L93_0354_6690.asc")
raster2 = RTClip(raster1, roi).run()

raster3 = STRasterize(buildings, raster2, attr="HAUTEUR").run()
# raster3 = STRasterize(buildings, raster2, attr=None).run()

plt.imshow(raster3.read(1), cmap="BrBG")
plt.show()
"""



'''
Created on 15 sept. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import re
import warnings
from zoneinfo import ZoneInfo

from pandas import read_csv
from t4gpd.commons.WarnUtils import WarnUtils
from t4gpd.io.AbstractReader import AbstractReader


class SensirionReader(AbstractReader):
    '''
    classdocs

    https://sensirion.com/products/catalog/SHT40/
    '''
    FIELD_NAMES = ['local_datetime', 'Tair', 'RH'] 

    RE1 = re.compile(r'^# EdfVersion=(\d\.\d)$')
    RE2 = re.compile(r'^# SensorFamily=(\w*).*$')
    RE3 = re.compile(r'^# SensorId=(.*)$')

    def __init__(self, inputFile, tzinfo="Europe/Paris"):
        '''
        Constructor
        '''
        warnings.formatwarning = WarnUtils.format_Warning
        self.inputFile = inputFile
        self.tzinfo = ZoneInfo(tzinfo)

    def __getEdfVersionAndSensorSpecs(self):
        with SensirionReader.opener(self.inputFile) as f:
            for nline, line in enumerate(f, start=1):
                line = line.strip()
                if (1 == nline):
                    version = float(self.RE1.search(line).group(1))
                elif (6 == nline):
                    sensorFamily = self.RE2.search(line).group(1)
                elif (7 == nline):
                    sensorId = self.RE3.search(line).group(1)
                    sensorId = sensorId.replace(':', '')
                    return version, sensorFamily, sensorId

    def run(self):
        version, sensorFamily, sensorId = self.__getEdfVersionAndSensorSpecs()

        if (4.0 != version):
            warnings.warn('EdfVersion is expected to be equal to 4.0!')

        # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html
        df = read_csv(SensirionReader.opener(self.inputFile), header=None,
                      names=self.FIELD_NAMES, parse_dates=['local_datetime'],
                      sep='\s+', skiprows=10, usecols=range(1, 4))
        # df.Epoch_UTC = df.Epoch_UTC.apply(lambda v: datetime.fromtimestamp(v))
        df.rename(columns={'local_datetime': 'timestamp'}, inplace=True)
        df['sensorFamily'] = sensorFamily
        df['sensorId'] = sensorId
        df['station'] = f'{sensorFamily}-{sensorId}'

        df.timestamp = df.timestamp.apply(lambda dt: dt.replace(tzinfo=self.tzinfo))
        if not df.timestamp.is_monotonic_increasing:
            msg = f"Timestamps in {self.inputFile} are not monotonically increasing"
            warnings.warn(msg)

        return df

'''
inputFile = '/home/tleduc/prj/nm-ilots-frais/terrain/220711/Sensirion_MyAmbience_SHT40_Gadget_0F0E_2022-07-11T18-49-50.435746.edf'
df = SensirionReader(inputFile).run()
print(df.head(5))
'''
